# pylint: disable-msg=too-many-lines
"""Contains the MachineController base class."""
import asyncio
import logging
import sys
import threading
from typing import Any, Callable, Dict, List, Set, Optional

from pkg_resources import iter_entry_points

from mpf._version import __version__
from mpf.core.clock import ClockBase
from mpf.core.config_validator import ConfigValidator
from mpf.core.data_manager import DataManager
from mpf.core.delays import DelayManager
from mpf.core.device_manager import DeviceCollection
from mpf.core.logging import LogMixin
from mpf.core.machine_vars import MachineVariables
from mpf.core.utility_functions import Util
from mpf.core.config_loader import MpfConfig

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.modes.game.code.game import Game   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.events import EventManager    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.switch_controller import SwitchController     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.show_controller import ShowController     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.service_controller import ServiceController   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.light_controller import LightController   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.platform_controller import PlatformController     # pylint: disable-msg=cyclic-import,unused-import

    from mpf.core.custom_code import CustomCode     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.mode_controller import ModeController     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.settings_controller import SettingsController     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.bcp.bcp import Bcp    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.text_ui import TextUi     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.assets.show import Show    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.switch import Switch   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.driver import Driver   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.mode import Mode  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.ball_device.ball_device import BallDevice  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.ball_controller import BallController     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.playfield import Playfield     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.placeholder_manager import PlaceholderManager     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.platforms.smart_virtual import SmartVirtualHardwarePlatform    # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.core.device_manager import DeviceManager   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.plugins.auditor import Auditor     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.light import Light     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.accelerometer import Accelerometer     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.drop_target import DropTarget, DropTargetBank  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.logic_blocks import Accrual, Sequence, Counter     # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.devices.servo import Servo     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.segment_display import SegmentDisplay      # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.shot_group import ShotGroup    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.shot import Shot   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.motor import Motor     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.digital_output import DigitalOutput    # pylint: disable-msg=cyclic-import,unused-import
    from logging import Logger  # noqa
    from mpf.devices.autofire import AutofireCoil   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.stepper import Stepper     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.config_players.show_player import ShowPlayer   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.dmd import Dmd     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.rgb_dmd import RgbDmd  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.flipper import Flipper     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.diverter import Diverter   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.multiball_lock import MultiballLock    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.multiball import Multiball     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.ball_hold import BallHold      # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.ball_save import BallSave      # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.magnet import Magnet           # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.state_machine import StateMachine  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.extra_ball import ExtraBall    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.extra_ball_group import ExtraBallGroup     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.achievement_group import AchievementGroup  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.achievement import Achievement     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.combo_switch import ComboSwitch    # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.score_queue import ScoreQueue      # pylint: disable-msg=cyclic-import,unused-import


# pylint: disable-msg=too-many-instance-attributes
class MachineController(LogMixin):

    """Base class for the Machine Controller object.

    The machine controller is the main entity of the entire framework. It's the
    main part that's in charge and makes things happen.

    Args:
    ----
        config(MpfConfig): The machine configuration
        options(dict): A dictionary of options built from the command line options
            used to launch mpf.py.
    """

    __slots__ = ["log", "options", "mpf_path", "machine_path", "_exception", "_boot_holds",
                 "is_init_done", "_done", "monitors", "plugins", "custom_code", "modes", "game",
                 "variables", "thread_stopper", "config", "config_validator",
                 "delay", "hardware_platforms", "default_platform", "clock",
                 "stop_future", "events", "switch_controller", "mode_controller", "settings",
                 "bcp", "ball_controller", "show_controller", "placeholder_manager", "device_manager", "auditor",
                 "tui", "service", "switches", "shows", "coils", "ball_devices", "lights", "playfield", "playfields",
                 "autofires", "_crash_handlers", "__dict__", "mpf_config", "is_shutting_down"]

    # pylint: disable-msg=too-many-statements
    def __init__(self, options: dict, config: MpfConfig) -> None:
        """Initialize machine controller."""
        super().__init__()
        self.log = logging.getLogger("Machine")     # type: Logger
        self.log.info("Mission Pinball Framework Core Engine v%s", __version__)
        self._crash_handlers = []   # type: List[Callable]
        self.is_shutting_down = False

        self.log.info("Command line arguments: %s", options)
        self.options = options

        self.mpf_path = config.get_mpf_path()
        self.log.info("MPF path: %s", self.mpf_path)

        self.machine_path = config.get_machine_path()
        self.log.info("Machine path: %s", self.machine_path)

        self.verify_system_info()
        self._exception = None      # type: Any
        self._boot_holds = set()    # type: Set[str]
        self.is_init_done = None    # type: Optional[asyncio.Event]

        self._done = False
        self.monitors = dict()      # type: Dict[str, Set[Callable]]
        self.plugins = list()       # type: List[Any]
        self.custom_code = list()   # type: List[CustomCode]
        self.modes = DeviceCollection(self, 'modes', None)          # type: Dict[str, Mode]
        self.game = None            # type: Optional[Game]
        self.thread_stopper = threading.Event()

        self.config = config.get_machine_config()       # type: Any
        self.mpf_config = config                        # type: MpfConfig
        self.config_validator = ConfigValidator(self, self.mpf_config.get_config_spec())

        self.variables = MachineVariables(self)  # type: MachineVariables

        # add some type hints
        if MYPY:   # pragma: no cover
            # controllers
            self.events = self.events                               # type: EventManager
            self.switch_controller = self.switch_controller         # type: SwitchController
            self.mode_controller = self.mode_controller             # type: ModeController
            self.settings = self.settings                           # type: SettingsController
            self.bcp = self.bcp                                     # type: Bcp
            self.ball_controller = self.ball_controller             # type: BallController
            self.show_controller = self.show_controller             # type: ShowController
            self.placeholder_manager = self.placeholder_manager     # type: PlaceholderManager
            self.device_manager = self.device_manager               # type: DeviceManager
            self.auditor = self.auditor                             # type: Auditor
            self.tui = self.tui                                     # type: TextUi
            self.service = self.service                             # type: ServiceController
            self.show_player = self.show_player                     # type: ShowPlayer
            self.light_controller = self.light_controller           # type: LightController
            self.platform_controller = self.platform_controller     # type: PlatformController

            # devices
            self.autofires = {}                         # type: Dict[str, AutofireCoil]
            self.motors = {}                            # type: Dict[str, Motor]
            self.digital_outputs = {}                   # type: Dict[str, DigitalOutput]
            self.shows = {}                             # type: Dict[str, Show]
            self.shots = {}                             # type: Dict[str, Shot]
            self.shot_groups = {}                       # type: Dict[str, ShotGroup]
            self.switches = {}                          # type: Dict[str, Switch]
            self.steppers = {}                          # type: Dict[str, Stepper]
            self.coils = {}                             # type: Dict[str, Driver]
            self.lights = {}                            # type: Dict[str, Light]
            self.ball_devices = {}                      # type: Dict[str, BallDevice]
            self.accelerometers = {}                    # type: Dict[str, Accelerometer]
            self.playfield = None                       # type: Optional[Playfield]
            self.playfields = {}                        # type: Dict[str, Playfield]
            self.counters = {}                          # type: Dict[str, Counter]
            self.sequences = {}                         # type: Dict[str, Sequence]
            self.accruals = {}                          # type: Dict[str, Accrual]
            self.drop_targets = {}                      # type: Dict[str, DropTarget]
            self.drop_target_banks = {}                 # type: Dict[str, DropTargetBank]
            self.servos = {}                            # type: Dict[str, Servo]
            self.segment_displays = {}                  # type: Dict[str, SegmentDisplay]
            self.dmds = {}                              # type: Dict[str, Dmd]
            self.rgb_dmds = {}                          # type: Dict[str, RgbDmd]
            self.flippers = {}                          # type: Dict[str, Flipper]
            self.diverters = {}                         # type: Dict[str, Diverter]
            self.multiball_locks = {}                   # type: Dict[str, MultiballLock]
            self.multiballs = {}                        # type: Dict[str, Multiball]
            self.ball_holds = {}                        # type: Dict[str, BallHold]
            self.ball_saves = {}                        # type: Dict[str, BallSave]
            self.magnets = {}                           # type: Dict[str, Magnet]
            self.state_machines = {}                    # type: Dict[str, StateMachine]
            self.extra_balls = {}                       # type: Dict[str, ExtraBall]
            self.extra_ball_groups = {}                 # type: Dict[str, ExtraBallGroup]
            self.achievements = {}                      # type: Dict[str, Achievement]
            self.achievement_groups = {}                # type: Dict[str, AchievementGroup]
            self.combo_switches = {}                    # type: Dict[str, ComboSwitch]
            self.score_queues = {}                      # type: Dict[str, ScoreQueue]

        self._set_machine_path()

        self.configure_logging(
            'Machine',
            self.config['logging']['console']['machine_controller'],
            self.config['logging']['file']['machine_controller'])

        self.delay = DelayManager(self)

        self.hardware_platforms = dict()    # type: Dict[str, SmartVirtualHardwarePlatform]
        self.default_platform = None        # type: Optional[SmartVirtualHardwarePlatform]

        self.clock = self._load_clock()
        self.stop_future = asyncio.Future()     # type: asyncio.Future

    def add_crash_handler(self, handler: Callable):
        """Add a crash handler which is called on a crash.

        This can be used to restore the output and prepare logging.
        """
        self._crash_handlers.append(handler)

    async def initialise_core_and_hardware(self) -> None:
        """Load core modules and hardware."""
        self._boot_holds = set()
        self.is_init_done = asyncio.Event()
        self.register_boot_hold('init')
        self._load_hardware_platforms()

        self._load_core_modules()
        # order is specified in mpfconfig.yaml

        self._validate_config()

        # This is called so hw platforms have a chance to register for events,
        # and/or anything else they need to do with core modules since
        # they're not set up yet when the hw platforms are constructed.
        await self._initialize_platforms()

    async def initialise(self) -> None:
        """Initialise machine."""
        await self.initialise_core_and_hardware()

        self._initialize_credit_string()

        self._register_config_players()
        self._register_system_events()
        self._load_machine_vars()
        await self._run_init_phases()
        self._init_phases_complete()

        await self._start_platforms()

        # wait until all boot holds were released
        assert self.is_init_done is not None
        await self.is_init_done.wait()
        await self.init_done()

    def _exception_handler(self, loop, context):    # pragma: no cover
        """Handle asyncio loop exceptions."""
        # call original exception handler
        loop.set_exception_handler(None)
        loop.call_exception_handler(context)

        # remember exception
        self._exception = context
        self.stop("Exception thrown")

    # pylint: disable-msg=no-self-use
    def _load_clock(self) -> ClockBase:  # pragma: no cover
        """Load clock and loop."""
        clock = ClockBase(self)
        clock.loop.set_exception_handler(self._exception_handler)
        return clock

    async def _run_init_phases(self) -> None:
        """Run init phases."""
        await self.events.post_queue_async("init_phase_1")
        '''event: init_phase_1

        desc: Posted during the initial boot up of MPF.
        '''
        await self.events.post_queue_async("init_phase_2")
        '''event: init_phase_2

        desc: Posted during the initial boot up of MPF.
        '''
        self._load_plugins()
        await self.events.post_queue_async("init_phase_3")
        '''event: init_phase_3

        desc: Posted during the initial boot up of MPF.
        '''
        self._load_custom_code()

        await self.events.post_queue_async("init_phase_4")
        '''event: init_phase_4

        desc: Posted during the initial boot up of MPF.
        '''

        await self.events.post_queue_async("init_phase_5")
        '''event: init_phase_5

        desc: Posted during the initial boot up of MPF.
        '''

    def _init_phases_complete(self, **kwargs) -> None:
        """Cleanup after init and remove boot holds."""
        del kwargs
        self.events.remove_all_handlers_for_event("init_phase_1")
        self.events.remove_all_handlers_for_event("init_phase_2")
        self.events.remove_all_handlers_for_event("init_phase_3")
        self.events.remove_all_handlers_for_event("init_phase_4")
        self.events.remove_all_handlers_for_event("init_phase_5")

        self.clear_boot_hold('init')

    async def _initialize_platforms(self) -> None:
        """Initialise all used hardware platforms."""
        init_done = []
        # collect all platform init futures
        for hardware_platform in list(self.hardware_platforms.values()):
            init_done.append(hardware_platform.initialize())

        # wait for all of them in parallel
        results = await asyncio.wait(init_done)
        for result in results[0]:
            result.result()

    async def _start_platforms(self) -> None:
        """Start all used hardware platforms."""
        for hardware_platform in list(self.hardware_platforms.values()):
            await hardware_platform.start()
            if not hardware_platform.features['tickless']:
                self.clock.schedule_interval(hardware_platform.tick, 1 / self.config['mpf']['default_platform_hz'])

    def _initialize_credit_string(self):
        """Set default credit string."""
        # Do this here so there's a credit_string var even if they're not using
        # the credits mode
        try:
            credit_string = self.config['credits']['free_play_string']
        except KeyError:
            credit_string = 'FREE PLAY'

        self.variables.set_machine_var('credits_string', credit_string)
        '''machine_var: credits_string

        desc: Holds a displayable string which shows how many
        credits are on the machine. For example, "CREDITS: 1". If the machine
        is set to free play, the value of this string will be "FREE PLAY".

        You can change the format and value of this string in the ``credits:``
        section of the machine config file.
        '''

    def _validate_config(self) -> None:
        """Validate game and machine config."""
        self.validate_machine_config_section('machine')
        self.validate_machine_config_section('game')
        self.validate_machine_config_section('mpf')

    def validate_machine_config_section(self, section: str) -> None:
        """Validate a config section."""
        if section not in self.config_validator.get_config_spec():
            return

        if section not in self.config:
            self.config[section] = dict()

        self.config[section] = self.config_validator.validate_config(
            section, self.config[section], section)

    def _register_system_events(self) -> None:
        """Register default event handlers."""
        self.events.add_handler('quit', self.stop)
        self.events.add_handler(self.config['mpf']['switch_tag_event'].
                                replace('%', 'quit'), self.stop)

    def _register_config_players(self) -> None:
        """Register config players."""
        # todo move this to config_player module
        for name, module_class in self.config['mpf']['config_players'].items():
            config_player_class = Util.string_to_class(module_class)
            setattr(self, '{}_player'.format(name),
                    config_player_class(self))

        self._register_plugin_config_players()

    def _register_plugin_config_players(self):
        """Register plugin config players."""
        self.debug_log("Registering Plugin Config Players")
        for entry_point in iter_entry_points(group='mpf.config_player',
                                             name=None):
            self.debug_log("Registering %s", entry_point)
            name, player = entry_point.load()(self)
            setattr(self, '{}_player'.format(name), player)

    def create_data_manager(self, config_name: str) -> DataManager:     # pragma: no cover
        """Return a new DataManager for a certain config.

        Args:
        ----
            config_name: Name of the config
        """
        return DataManager(self, config_name)

    def _load_machine_vars(self) -> None:
        """Load machine vars from data manager."""
        machine_var_data_manager = self.create_data_manager('machine_vars')
        current_time = self.clock.get_time()

        self.variables.load_machine_vars(machine_var_data_manager, current_time)

    def _set_machine_path(self) -> None:
        """Add the machine folder to sys.path so we can import modules from it."""
        sys.path.insert(0, self.machine_path)

    def verify_system_info(self):
        """Dump information about the Python installation to the log.

        Information includes Python version, Python executable, platform, and
        core architecture.
        """
        python_version_info = sys.version_info

        if not (python_version_info[0] == 3 and python_version_info[1] in (5, 6, 7, 8, 9)):
            raise AssertionError("Incorrect Python version. MPF requires "
                                 "Python 3.5, 3.6, 3.7, 3.8 or 3.9. You have Python {}.{}.{}."
                                 .format(python_version_info[0], python_version_info[1],
                                         python_version_info[2]))

        self.log.info("Platform: %s", sys.platform)
        self.log.info("Python executable location: %s", sys.executable)

        if sys.maxsize < 2**32:
            self.log.info("Python version: %s.%s.%s (32-bit)", python_version_info[0],
                          python_version_info[1], python_version_info[2])
        else:
            self.log.info("Python version: %s.%s.%s (64-bit)", python_version_info[0],
                          python_version_info[1], python_version_info[2])

    def _load_core_modules(self) -> None:
        """Load core modules."""
        self.debug_log("Loading core modules...")
        for name, module_class in self.config['mpf']['core_modules'].items():
            self.debug_log("Loading '%s' core module", module_class)
            m = Util.string_to_class(module_class)(self)
            setattr(self, name, m)

    def _load_hardware_platforms(self) -> None:
        """Load all hardware platforms."""
        self.validate_machine_config_section('hardware')

        # load internal platforms
        self.add_platform("drivers")

        # if platform is forced use that one
        if self.options['force_platform']:
            self.add_platform(self.options['force_platform'])
            self.set_default_platform(self.options['force_platform'])
            return

        # otherwise load all platforms
        for section, platforms in self.config['hardware'].items():
            if section == 'driverboards':
                continue
            for hardware_platform in platforms:
                if hardware_platform.lower() != 'default':
                    self.add_platform(hardware_platform)

        # set default platform
        self.set_default_platform(self.config['hardware']['platform'][0])

    def _load_plugins(self) -> None:
        """Load plugins."""
        self.debug_log("Loading plugins...")

        for plugin in self.config['mpf']['plugins']:

            self.debug_log("Loading '%s' plugin", plugin)

            plugin_obj = Util.string_to_class(plugin)(self)
            self.plugins.append(plugin_obj)

    def _load_custom_code(self) -> None:
        """Load custom code."""
        if 'scriptlets' in self.config:
            self.debug_log("Loading scriptlets (deprecated).")
            for scriptlet in Util.string_to_event_list(self.config['scriptlets']):
                self.debug_log("Loading '%s' scriptlet (deprecated)", scriptlet)
                scriptlet_obj = Util.string_to_class(self.config['mpf']['paths']['scriptlets'] + "." + scriptlet)(
                    machine=self,
                    name=scriptlet.split('.')[1])
                self.custom_code.append(scriptlet_obj)

        if 'custom_code' in self.config:
            self.debug_log("Loading custom code.")

            for custom_code in Util.string_to_event_list(self.config['custom_code']):

                self.debug_log("Loading '%s' custom code", custom_code)

                custom_code_obj = Util.string_to_class(custom_code)(
                    machine=self,
                    name=custom_code)

                self.custom_code.append(custom_code_obj)

    async def reset(self) -> None:
        """Reset the machine.

        This method is safe to call. It essentially sets up everything from
        scratch without reloading the config files and assets from disk. This
        method is called after a game ends and before attract mode begins.
        """
        self.debug_log('Resetting...')

        await self.events.post_queue_async('machine_reset_phase_1')
        '''Event: machine_reset_phase_1

        Desc: The first phase of resetting the machine.

        These events are posted when MPF boots (after the init_phase events are
        posted), and they're also posted subsequently when the machine is reset
        (after existing the service mode, for example).

        This is a queue event. The machine reset phase 1 will not be complete
        until the queue is cleared.

        '''

        await self.events.post_queue_async('machine_reset_phase_2')
        '''Event: machine_reset_phase_2

        Desc: The second phase of resetting the machine.

        These events are posted when MPF boots (after the init_phase events are
        posted), and they're also posted subsequently when the machine is reset
        (after existing the service mode, for example).

        This is a queue event. The machine reset phase 2 will not be complete
        until the queue is cleared.

        '''

        await self.events.post_queue_async('machine_reset_phase_3')
        '''Event: machine_reset_phase_3

        Desc: The third phase of resetting the machine.

        These events are posted when MPF boots (after the init_phase events are
        posted), and they're also posted subsequently when the machine is reset
        (after existing the service mode, for example).

        This is a queue event. The machine reset phase 3 will not be complete
        until the queue is cleared.

        '''

        """Called when the machine reset process is complete."""
        self.debug_log('Reset Complete')
        await self.events.post_async('reset_complete')
        '''event: reset_complete

        desc: The machine reset process is complete

        '''

    def add_platform(self, name: str) -> None:
        """Make an additional hardware platform interface available to MPF.

        Args:
        ----
            name: String name of the platform to add. Must match the name of a
                platform file in the mpf/platforms folder (without the .py
                extension).
        """
        if name not in self.hardware_platforms:
            if name in self.config['mpf']['platforms']:
                # if platform is in config load it
                try:
                    hardware_platform = Util.string_to_class(self.config['mpf']['platforms'][name])
                except ImportError as e:  # pragma: no cover
                    if e.name != name:  # do not swallow unrelated errors
                        raise
                    raise ImportError("Cannot add hardware platform {}. This is "
                                      "not a valid platform name".format(name))

            else:
                # check entry points
                entry_points = list(iter_entry_points(group='mpf.platforms', name=name))
                if entry_points:
                    # load platform from entry point
                    self.debug_log("Loading platform %s from external entry_point", name)
                    if len(entry_points) != 1:
                        raise AssertionError("Got more than one entry point for platform {}: {}".format(name,
                                                                                                        entry_points))

                    hardware_platform = entry_points[0].load()
                else:
                    raise AssertionError("Unknown platform {}".format(name))

            self.hardware_platforms[name] = hardware_platform(self)

    def set_default_platform(self, name: str) -> None:
        """Set the default platform.

        It is used if a device class-specific or device-specific platform is not specified.

        Args:
        ----
            name: String name of the platform to set to default.
        """
        try:
            self.default_platform = self.hardware_platforms[name]
            self.debug_log("Setting default platform to '%s'", name)
        except KeyError:
            raise AssertionError("Cannot set default platform to '{}', as that's not"
                                 " a currently active platform".format(name))

    def register_monitor(self, monitor_class: str, monitor: Callable[..., Any]) -> None:
        """Register a monitor.

        Args:
        ----
            monitor_class: String name of the monitor class for this monitor
                that's being registered.
            monitor: Callback to notify

        MPF uses monitors to allow components to monitor certain internal
        elements of MPF.

        For example, a player variable monitor could be setup to be notified of
        any changes to a player variable, or a switch monitor could be used to
        allow a plugin to be notified of any changes to any switches.

        The MachineController's list of registered monitors doesn't actually
        do anything. Rather it's a dictionary of sets which the monitors
        themselves can reference when they need to do something. We just needed
        a central registry of monitors.

        """
        if monitor_class not in self.monitors:
            self.monitors[monitor_class] = set()

        self.monitors[monitor_class].add(monitor)

    def initialise_mpf(self):
        """Initialise MPF."""
        self.info_log("Initialise MPF.")
        timeout = 30 if self.options["production"] else None
        try:
            init = asyncio.ensure_future(self.initialise())
            self.clock.loop.run_until_complete(Util.first([init, self.stop_future], cancel_others=False,
                                                          timeout=timeout))
        except asyncio.TimeoutError:
            self._crash_shutdown()
            self.error_log("MPF needed more than {}s for initialisation. Aborting!".format(timeout))
            return False
        except RuntimeError:
            self._crash_shutdown()
            # do not show a runtime useless runtime error
            self.error_log("Failed to initialise MPF")
            return False
        if init.done() and init.exception():
            self._crash_shutdown()
            try:
                raise init.exception()
            except:     # noqa
                self.log.exception("Failed to initialise MPF")
            return False

        return True

    def run(self) -> None:
        """Start the main machine run loop."""
        if not self.initialise_mpf():
            return

        self.info_log("Starting the main run loop.")
        self._run_loop()

    def stop(self, reason=None, **kwargs) -> None:
        """Perform a graceful exit of MPF."""
        del kwargs
        if self.stop_future.done():
            return

        self.clock.loop.call_soon_threadsafe(self._stop_loop, reason)

    def _stop_loop(self, reason):
        self.stop_future.set_result(reason)

    def _do_stop(self) -> None:
        self.is_shutting_down = True
        self.log.info("Shutting down...")
        self.events.post('shutdown')
        '''event: shutdown
        desc: Posted when the machine is shutting down to give all modules a
        chance to shut down gracefully.

        '''

        self.events.process_event_queue()
        self.shutdown()

    def _crash_shutdown(self):
        """MPF crashed. Cleanup as good as we can."""
        # call crash handlers
        for handler in self._crash_handlers:
            handler()
        if hasattr(self, "events") and hasattr(self, "clock"):
            # if we already got events and a clock use normal shutdown
            self._do_stop()
        else:
            # otherwise just shutdown
            self.shutdown()

    def shutdown(self) -> None:
        """Shutdown the machine."""
        self.thread_stopper.set()
        if hasattr(self, "device_manager"):
            self.device_manager.stop_devices()
        self._platform_stop()

        self.clock.loop.stop()
        # this is needed to properly close all sockets
        self.clock.loop.run_forever()
        self.clock.loop.close()

    def _run_loop(self) -> None:    # pragma: no cover
        # Main machine run loop with when the default platform interface
        # specifies the MPF should control the main timer

        try:
            reason = self.clock.run(self.stop_future)
        except KeyboardInterrupt:
            print("Shutdown because of keyboard interrupts")
            return
        except BaseException:   # pylint: disable-msg=broad-except
            # this happens when receiving a signal
            self.log.exception("Loop exited with exception")
            return

        if self._exception:
            self._crash_shutdown()

            print("Shutdown because of an exception:")
            try:
                raise self._exception['exception']
            except:     # noqa
                self.log.exception("Runtime Exception")
        else:
            self._do_stop()
            print("Shutdown reason: {}".format(reason))

    def _platform_stop(self) -> None:
        """Stop all platforms."""
        for hardware_platform in list(self.hardware_platforms.values()):
            hardware_platform.stop()

    def get_platform_sections(self, platform_section: str, overwrite: str) -> "SmartVirtualHardwarePlatform":
        """Return platform section."""
        assert self.default_platform is not None
        if overwrite == "drivers":
            return self.hardware_platforms[overwrite]

        if self.options['force_platform']:
            return self.default_platform

        if not overwrite:
            if self.config['hardware'][platform_section][0] != 'default':
                return self.hardware_platforms[self.config['hardware'][platform_section][0]]

            return self.default_platform

        try:
            return self.hardware_platforms[overwrite]
        except KeyError:
            raise AssertionError("Platform \"{}\" has not been loaded. Please add it to your \"hardware\" section.".
                                 format(overwrite))

    def register_boot_hold(self, hold: str) -> None:
        """Register a boot hold."""
        assert self.is_init_done is not None
        if self.is_init_done.is_set():
            raise AssertionError("Register hold after init_done")
        self._boot_holds.add(hold)

    def clear_boot_hold(self, hold: str) -> None:
        """Clear a boot hold."""
        assert self.is_init_done is not None
        if self.is_init_done.is_set():
            raise AssertionError("Clearing hold after init_done")
        self._boot_holds.remove(hold)
        self.debug_log('Clearing boot hold %s. Holds remaining: %s', hold, self._boot_holds)
        if not self._boot_holds:
            self.is_init_done.set()

    async def init_done(self) -> None:
        """Finish init.

        Called when init is done and all boot holds are cleared.
        """
        await self.events.post_async("init_done")
        '''event: init_done

        desc: Posted when the initial (one-time / boot) init phase is done. In
        other words, once this is posted, MPF is booted and ready to go.
        '''

        await self.reset()



"""Contains the drivers and interface code for pinball machines which use the Multimorphic R-ROC hardware controllers.

This code can be used with P-ROC driver boards, or with Stern SAM, Stern
Whitestar, Williams WPC, or Williams WPC95 driver boards.

Much of this code is from the P-ROC drivers section of the pyprocgame project,
written by Adam Preble and Gerry Stellenberg. It was originally released under
the MIT license and is released here under the MIT License.

More info on the P-ROC hardware platform: http://pinballcontrollers.com/

Original code source on which this module was based:
https://github.com/preble/pyprocgame
"""
from typing import Dict

from mpf.core.platform import DmdPlatform, DriverConfig, SwitchConfig, SegmentDisplaySoftwareFlashPlatform
from mpf.platforms.interfaces.dmd_platform import DmdPlatformInterface
from mpf.platforms.interfaces.segment_display_platform_interface import SegmentDisplaySoftwareFlashPlatformInterface
from mpf.platforms.p_roc_common import PDBConfig, PROCBasePlatform
from mpf.core.utility_functions import Util
from mpf.platforms.p_roc_devices import PROCDriver

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


class PRocHardwarePlatform(PROCBasePlatform, DmdPlatform, SegmentDisplaySoftwareFlashPlatform):

    """Platform class for the P-ROC hardware controller.

    Args:
    ----
        machine: The MachineController instance.
    """

    __slots__ = ["dmd", "alpha_display", "aux_port", "_use_extended_matrix",
                 "_use_first_eight_direct_inputs"]

    def __init__(self, machine):
        """Initialise P-ROC."""
        super().__init__(machine)
        # validate config for p_roc
        self.config = self.machine.config_validator.validate_config("p_roc", self.machine.config.get('p_roc', {}))
        self._configure_device_logging_and_debug('P-Roc', self.config)

        if self.config['driverboards']:
            self.machine_type = self.pinproc.normalize_machine_type(self.config['driverboards'])
        else:
            self.machine_type = self.pinproc.normalize_machine_type(self.machine.config['hardware']['driverboards'])

        self.dmd = None
        self.alpha_display = None
        self.aux_port = None

        self._use_extended_matrix = False
        self._use_first_eight_direct_inputs = False

    async def connect(self):
        """Connect to the P-Roc."""
        await super().connect()

        self.aux_port = AuxPort(self)
        self.aux_port.reset()

        # Because PDBs can be configured in many different ways, we need to
        # traverse the YAML settings to see how many PDBs are being used.
        # Then we can configure the P-ROC appropriately to use those PDBs.
        # Only then can we relate the YAML coil/light #'s to P-ROC numbers for
        # the collections.
        if self.machine_type == self.pinproc.MachineTypePDB:
            self.debug_log("Configuring P-ROC for PDBs (P-ROC driver boards)")
            self.pdbconfig = PDBConfig(self, self.machine.config, self.pinproc.DriverCount)

        else:
            self.debug_log("Configuring P-ROC for OEM driver boards")

    def _get_default_subtype(self):
        """Return default subtype for P-Roc."""
        return "matrix"

    def __repr__(self):
        """Return string representation."""
        return '<Platform.P-ROC>'

    def get_info_string(self):
        """Dump infos about boards."""
        infos = "Firmware Version: {} Firmware Revision: {} Hardware Board ID: {}\n".format(
            self.version, self.revision, self.hardware_version)
        return infos

    @classmethod
    def get_coil_config_section(cls):
        """Return coil config section."""
        return "p_roc_coils"

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict):
        """Create a P-ROC driver.

        Typically drivers are coils or flashers, but for the P-ROC this is
        also used for matrix-based lights.

        Args:
        ----
            config: Dictionary of settings for the driver.
            number: Number of this driver
            platform_settings: Platform specific setting for this driver.

        Returns a reference to the PROCDriver object which is the actual object
        you can use to pulse(), patter(), enable(), etc.

        """
        # todo need to add Aux Bus support
        # todo need to add virtual driver support for driver counts > 256

        # Find the P-ROC number for each driver. For P-ROC driver boards, the
        # P-ROC number is specified via the Ax-By-C format. For OEM driver
        # boards configured via driver numbers, libpinproc's decode() method
        # can provide the number.

        if self.machine_type == self.pinproc.MachineTypePDB:
            proc_num = self.pdbconfig.get_proc_coil_number(str(number))
            if proc_num == -1:
                raise AssertionError("Driver {} cannot be controlled by the P-ROC. ".format(str(number)))
        else:
            proc_num = self.pinproc.decode(self.machine_type, str(number))

        polarity = platform_settings.get("polarity", None)
        driver = PROCDriver(proc_num, config, self, number, polarity)
        self._late_init_futures.append(driver.initialise())

        return driver

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict):
        """Configure a P-ROC switch.

        Args:
        ----
            number: String number of the switch to configure.
            config: SwitchConfig settings.
            platform_config: Platform specific settings.

        Returns: A configured switch object.

        """
        del platform_config
        try:
            if number.startswith("SD") and 0 <= int(number[2:]) <= 7:
                self._use_first_eight_direct_inputs = True
            _, y = number.split('/', 2)
            if int(y) > 7:
                self._use_extended_matrix = True
        except ValueError:
            pass

        if self._use_extended_matrix and self._use_first_eight_direct_inputs:
            raise AssertionError(
                "P-Roc vannot use extended matrix and the first eight direct inputs at the same "
                "time. Either only use SD8 to SD31 or only use matrix X/Y with Y <= 7. Offending "
                "switch: {}".format(number))

        if self.machine_type == self.pinproc.MachineTypePDB:
            proc_num = self.pdbconfig.get_proc_switch_number(str(number))
            if proc_num == -1:
                raise AssertionError("Switch {} cannot be controlled by the P-ROC. ".format(str(number)))

        else:
            proc_num = self.pinproc.decode(self.machine_type, str(number))
        return self._configure_switch(config, proc_num)

    async def get_hw_switch_states(self) -> Dict[str, bool]:
        """Read in and set the initial switch state.

        The P-ROC uses the following values for hw switch states:
        1 - closed (debounced)
        2 - open (debounced)
        3 - closed (not debounced)
        4 - open (not debounced)
        """
        switch_states = await self.run_proc_cmd("switch_get_states")
        states = {}

        for switch, state in enumerate(switch_states):
            states[switch] = bool(state in (1, 3))

        return states

    def configure_dmd(self):
        """Configure a hardware DMD connected to a classic P-ROC."""
        self.dmd = PROCDMD(self, self.machine)
        return self.dmd

    async def configure_segment_display(self, number: str, platform_settings) \
            -> "SegmentDisplaySoftwareFlashPlatformInterface":
        """Configure display."""
        del platform_settings
        number_int = int(number)
        if 0 < number_int >= 4:
            raise AssertionError("Number must be between 0 and 3 for p_roc segment display.")

        if not self.alpha_display:
            self.alpha_display = AuxAlphanumericDisplay(self, self.aux_port)

        display = PRocAlphanumericDisplay(self.alpha_display, number_int)
        self._handle_software_flash(display)
        return display

    def process_events(self, events):
        """Process events from the P-Roc."""
        for event in events:
            event_type = event['type']
            event_value = event['value']
            if event_type == self.pinproc.EventTypeDMDFrameDisplayed:
                pass
            elif event_type == self.pinproc.EventTypeSwitchClosedDebounced:
                self.machine.switch_controller.process_switch_by_num(
                    state=1, num=event_value, platform=self)
            elif event_type == self.pinproc.EventTypeSwitchOpenDebounced:
                self.machine.switch_controller.process_switch_by_num(
                    state=0, num=event_value, platform=self)
            elif event_type == self.pinproc.EventTypeSwitchClosedNondebounced:
                self.machine.switch_controller.process_switch_by_num(
                    state=1, num=event_value, platform=self)
            elif event_type == self.pinproc.EventTypeSwitchOpenNondebounced:
                self.machine.switch_controller.process_switch_by_num(
                    state=0, num=event_value, platform=self)
            else:
                self.log.warning("Received unrecognized event from the P-ROC. "
                                 "Type: %s, Value: %s", event_type, event_value)


class PROCDMD(DmdPlatformInterface):

    """Parent class for a physical DMD attached to a P-ROC.

    Args:
    ----
        platform: Reference to the MachineController's proc attribute.
        machine: Reference to the MachineController
    """

    __slots__ = ["machine", "platform"]

    def __init__(self, platform, machine):
        """Set up DMD."""
        self.platform = platform        # type: PROCBasePlatform
        self.machine = machine          # type: MachineController

        # dmd_timing defaults should be 250, 400, 180, 800
        if self.machine.config['p_roc']['dmd_timing_cycles']:
            dmd_timing = Util.string_to_event_list(
                self.machine.config['p_roc']['dmd_timing_cycles'])

            self.platform.run_proc_cmd_no_wait("dmd_update_config", dmd_timing)

    def set_brightness(self, brightness: float):
        """Set brightness."""
        # currently not supported. can be implemented using dmd_timing_cycles
        assert brightness == 1.0

    def update(self, data):
        """Update the DMD with a new frame.

        Args:
        ----
            data: A 4096-byte raw string.

        """
        if len(data) == 4096:
            self.platform.run_proc_cmd_no_wait("_dmd_send", data)
        else:
            self.machine.log.warning("Received DMD frame of length %s instead"
                                     "of 4096. Discarding...", len(data))


class AuxPort:

    """Aux port on the P-Roc."""

    __slots__ = ["platform", "_commands"]

    def __init__(self, platform):
        """Initialise aux port."""
        self.platform = platform
        self._commands = []

    def reset(self):
        """Reset aux port."""
        commands = [self.platform.pinproc.aux_command_disable()]

        for _ in range(1, 255):
            commands += [self.platform.pinproc.aux_command_jump(0)]

        self.platform.run_proc_cmd_no_wait("aux_send_commands", 0, commands)

    def reserve_index(self):
        """Return index of next free command slot and reserve it."""
        self._commands += [[]]
        return len(self._commands) - 1

    def update(self, index, commands):
        """Update command slot with command."""
        self._commands[index] = commands
        self._write_commands()

    def _write_commands(self):
        """Write commands to hardware."""
        # disable program
        commands = [self.platform.pinproc.aux_command_disable()]
        # build command list
        for command_set in self._commands:
            commands += command_set
            self.platform.run_proc_cmd_no_wait("aux_send_commands", 0, commands)

        # jump from slot 0 to slot 1. overwrites the disable
            self.platform.run_proc_cmd_no_wait("aux_send_commands", 0, [self.platform.pinproc.aux_command_jump(1)])


class PRocAlphanumericDisplay(SegmentDisplaySoftwareFlashPlatformInterface):

    """Since AuxAlphanumericDisplay updates all four displays wrap it and set the correct offset."""

    __slots__ = ["display"]

    def __init__(self, display, index):
        """Initialise alpha numeric display."""
        super().__init__(index)
        self.display = display

    def _set_text(self, text: str):
        """Set digits to display."""
        self.display.set_text(text, self.number)


class AuxAlphanumericDisplay:

    """An alpha numeric display connected to the aux port on the P-Roc."""

    # Start at ASCII table offset 32: ' '
    ascii_segments = [0x0000,  # ' '
                      0x016a,  # '!' Random Debris Character 1
                      0x3014,  # '"' Random Debris Character 2
                      0x5d80,  # '#' Random Debris Character 3
                      0x00a4,  # '$' Random Debris Character 4
                      0x3270,  # '%' Random Debris Character 5
                      0x4640,  # '&' Random Debris Character 6
                      0x0200,  # '''
                      0x1400,  # '('
                      0x4100,  # ')'
                      0x7f40,  # '*'
                      0x2a40,  # '+'
                      0x8080,  # ','
                      0x0840,  # '-'
                      0x8000,  # '.'
                      0x4400,  # '/'

                      0x003f,  # '0'
                      0x0006,  # '1'
                      0x085b,  # '2'
                      0x084f,  # '3'
                      0x0866,  # '4'
                      0x086d,  # '5'
                      0x087d,  # '6'
                      0x0007,  # '7'
                      0x087f,  # '8'
                      0x086f,  # '9'

                      0x0821,  # ':' Random Debris Character 7
                      0x1004,  # ';' Random Debris Character 8
                      0x1c00,  # '<' Left Arrow
                      0x1386,  # '=' Random Debris Character 9
                      0x4140,  # '>' Right Arrow
                      0x0045,  # '?' Random Debris Character 10
                      0x4820,  # '@' Random Debris Character 11

                      0x0877,  # 'A'
                      0x2a4f,  # 'B'
                      0x0039,  # 'C'
                      0x220f,  # 'D'
                      0x0879,  # 'E'
                      0x0871,  # 'F'
                      0x083d,  # 'G'
                      0x0876,  # 'H'
                      0x2209,  # 'I'
                      0x001e,  # 'J'
                      0x1470,  # 'K'
                      0x0038,  # 'L'
                      0x0536,  # 'M'
                      0x1136,  # 'N'
                      0x003f,  # 'O'
                      0x0873,  # 'P'
                      0x103f,  # 'Q'
                      0x1873,  # 'R'
                      0x086d,  # 'S'
                      0x2201,  # 'T'
                      0x003e,  # 'U'
                      0x4430,  # 'V'
                      0x5036,  # 'W'
                      0x5500,  # 'X'
                      0x2500,  # 'Y'
                      0x4409,  # 'Z'

                      0x6004,  # '[' Random Debris Character 12
                      0x6411,  # '\' Random Debris Character 13
                      0x780a,  # ']' Random Debris Character 14
                      0x093a,  # '^' Random Debris Character 15
                      0x0008,  # '_'
                      0x2220,  # '`' Random Debris Character 16

                      0x0c56,  # 'a' Broken Letter a
                      0x684e,  # 'b' Broken Letter b
                      0x081c,  # 'c' Broken Letter c
                      0x380e,  # 'd' Broken Letter d
                      0x1178,  # 'e' Broken Letter e
                      0x4831,  # 'f' Broken Letter f
                      0x083d,  # 'g' Broken Letter g NOT CREATED YET
                      0x0854,  # 'h' Broken Letter h
                      0x2209,  # 'i' Broken Letter i NOT CREATED YET
                      0x001e,  # 'j' Broken Letter j NOT CREATED YET
                      0x1070,  # 'k' Broken Letter k
                      0x0038,  # 'l' Broken Letter l NOT CREATED YET
                      0x0536,  # 'm' Broken Letter m NOT CREATED YET
                      0x1136,  # 'n' Broken Letter n NOT CREATED YET
                      0x085c,  # 'o' Broken Letter o
                      0x0873,  # 'p' Broken Letter p NOT CREATED YET
                      0x103f,  # 'q' Broken Letter q NOT CREATED YET
                      0x1c72,  # 'r' Broken Letter r
                      0x116c,  # 's' Broken Letter s
                      0x2120,  # 't' Broken Letter t
                      0x003e,  # 'u' Broken Letter u NOT CREATED YET
                      0x4430,  # 'v' Broken Letter v NOT CREATED YET
                      0x5036,  # 'w' Broken Letter w NOT CREATED YET
                      0x5500,  # 'x' Broken Letter x NOT CREATED YET
                      0x2500,  # 'y' Broken Letter y NOT CREATED YET
                      0x4409  # 'z' Broken Letter z NOT CREATED YET
                      ]

    strobes = [8, 9, 10, 11, 12]
    full_intensity_delay = 350  # microseconds
    inter_char_delay = 40       # microseconds

    __slots__ = ["platform", "aux_controller", "aux_index", "texts"]

    def __init__(self, platform, aux_controller):
        """Initialise the alphanumeric display."""
        self.platform = platform
        self.aux_controller = aux_controller
        self.aux_index = aux_controller.reserve_index()
        self.texts = ["        "] * 4

    def set_text(self, text, index):
        """Set text for display."""
        if len(text) != 8:
            text = text[0:8].rjust(8, ' ')
        self.texts[index] = text

        # build expected format
        input_strings = [self.texts[0] + self.texts[1], self.texts[2] + self.texts[3]]
        self.display(input_strings)

    def display(self, input_strings, intensities=None):
        """Set display text."""
        strings = []

        if intensities is None:
            intensities = [[1] * 16] * 2

        # Make sure strings are at least 16 chars.
        # Then convert each string to a list of chars.
        for j in range(0, 2):
            input_strings[j] = input_strings[j]
            if len(input_strings[j]) < 16:
                input_strings[j] += ' ' * (16 - len(input_strings[j]))
            strings += [list(input_strings[j])]

        # Make sure insensities are 1 or less
        for i in range(0, 16):
            for j in range(0, 2):
                if intensities[j][i] > 1:
                    intensities[j][i] = 1

        commands = []
        char_on_time = []
        char_off_time = []

        # Initialize a 2x16 array for segments value
        segs = [[0] * 16 for _ in range(2)]

        # Loop through each character
        for i in range(0, 16):

            # Activate the character position (this goes to both displayas)
            commands += [self.platform.pinproc.aux_command_output_custom(i, 0, self.strobes[0], False, 0)]

            for j in range(0, 2):
                segs[j][i] = self.ascii_segments[ord(strings[j][i]) - 32]

                # Check for commas or periods.
                # If found, squeeze comma into previous character.
                # No point checking the last character (plus, this avoids an
                # indexing error by not checking i+1 on the 16th char.
                if i < 15:
                    comma_dot = strings[j][i + 1]
                    if comma_dot in (".", ","):
                        segs[j][i] |= self.ascii_segments[ord(comma_dot) - 32]
                        strings[j].remove(comma_dot)
                        # Append a space to ensure there are enough chars.
                        strings[j].append(' ')
                # character is 16 bits long, characters are loaded in 2 lots of 8 bits,
                # for each display (4 enable lines total)
                commands += [self.platform.pinproc.aux_command_output_custom(
                    segs[j][i] & 0xff, 0,
                    self.strobes[j * 2 + 1], False, 0)]     # first 8 bits of characater data
                commands += [self.platform.pinproc.aux_command_output_custom(
                    (segs[j][i] >> 8) & 0xff, 0,
                    self.strobes[j * 2 + 2], False, 0)]     # second 8 bits of characater data

                char_on_time += [intensities[j][i] * self.full_intensity_delay]
                char_off_time += [self.inter_char_delay + (self.full_intensity_delay - char_on_time[j])]

            if char_on_time[0] < char_on_time[1]:
                first = 0
                second = 1
            else:
                first = 1
                second = 0

            # Determine amount of time to leave the other char on after the
            # first is off.
            between_delay = char_on_time[second] - char_on_time[first]

            # Not sure if the hardware will like a delay of 0
            # Use 2 to be extra safe.  2 microseconds won't affect display.
            if between_delay == 0:
                between_delay = 2

            # Delay until it's time to turn off the character with the lowest intensity
            commands += [self.platform.pinproc.aux_command_delay(char_on_time[first])]
            commands += [self.platform.pinproc.aux_command_output_custom(0, 0, self.strobes[first * 2 + 1], False, 0)]
            commands += [self.platform.pinproc.aux_command_output_custom(0, 0, self.strobes[first * 2 + 2], False, 0)]

            # Delay until it's time to turn off the other character.
            commands += [self.platform.pinproc.aux_command_delay(between_delay)]
            commands += [self.platform.pinproc.aux_command_output_custom(0, 0, self.strobes[second * 2 + 1], False, 0)]
            commands += [self.platform.pinproc.aux_command_output_custom(0, 0, self.strobes[second * 2 + 2], False, 0)]

            # Delay for the inter-digit delay.
            commands += [self.platform.pinproc.aux_command_delay(char_off_time[second])]

        # Send the new list of commands to the Aux port controller.
        self.aux_controller.update(self.aux_index, commands)



"""Queue Relay Config Player."""
from typing import List

from mpf.core.utility_functions import Util

from mpf.core.config_player import ConfigPlayer


class QueueRelayPlayer(ConfigPlayer):

    """Blocks queue events and converts them to normal events."""

    config_file_section = 'queue_relay_player'

    __slots__ = []  # type: List[str]

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Block queue event."""
        del calling_context
        # prevent reposting _from_bcp setting
        kwargs.pop("_from_bcp", None)
        try:
            queue = kwargs.pop('queue')
        except KeyError:
            raise AssertionError(
                "Can only use queue relay player with queue event.")

        instance_dict = self._get_instance_dict(context)

        p = settings['priority']
        if priority:
            p += priority
        handler = self.machine.events.add_handler(settings['wait_for'],
                                                  self._callback, p,
                                                  context=context,
                                                  queue=queue)
        instance_dict[queue] = handler
        queue.wait()
        if settings["pass_args"] and kwargs and settings['args']:
            args = Util.dict_merge(kwargs, settings['args'])
        elif settings["args"]:
            args = settings["args"]
        elif settings["pass_args"]:
            args = kwargs
        else:
            args = {}

        self.machine.events.post(settings['post'], **args)

    def clear_context(self, context):
        """Clear all queues and remove handlers."""
        for queue, handler in self._get_instance_dict(context).items():
            self.machine.events.remove_handler_by_key(handler)
            queue.clear()

        self._reset_instance_dict(context)

    def validate_config_entry(self, settings, name):
        """Validate one entry of this player."""
        config = self._parse_config(settings, name)
        return config

    def _callback(self, queue, context, **kwargs):
        del kwargs
        instance_dict = self._get_instance_dict(context)
        # bail out on error
        if queue not in instance_dict:
            raise AssertionError("Queue {} missing in instance dict: {}.".format(queue, instance_dict))

        self.machine.events.remove_handler_by_key(instance_dict[queue])
        del instance_dict[queue]
        queue.clear()

    def get_express_config(self, value):
        """No express config."""
        raise AssertionError("Not supported")



"""Test multiballs and multiball_locks."""
from mpf.tests.MpfGameTestCase import MpfGameTestCase


class TestMultiBall(MpfGameTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/multiball/'

    def get_platform(self):
        return 'smart_virtual'

    def testSimpleMultiball(self):
        self.mock_event("multiball_mb1_ended")
        self.mock_event("multiball_mb1_ball_lost")

        # prepare game
        self.fill_troughs()
        self.assertFalse(self.machine.multiballs["mb1"].enabled)

        # start game
        self.start_game()

        # mb1 should not start because its not enabled
        self.post_event("mb1_start")
        self.assertEqual(0, self.machine.multiballs["mb1"].balls_added_live)
        self.assertFalse(self.machine.multiballs["mb1"].enabled)

        self.post_event("mb1_enable")

        # multiball should be enabled now but not started
        self.assertTrue(self.machine.multiballs["mb1"].enabled)
        self.assertEqual(0, self.machine.multiballs["mb1"].balls_added_live)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # ball drains right away
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # multiball not started. game should end
        self.assertEqual(None, self.machine.game)

        # start game again
        self.start_game()
        self.post_event("mb1_enable")

        # multiball should be enabled
        self.assertTrue(self.machine.multiballs["mb1"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        self.post_event("mb1_disable")
        # nothing happens
        self.post_event("mb1_start")
        self.assertEqual(0, self.machine.multiballs["mb1"].balls_added_live)

        # mb start
        self.post_event("mb1_enable")
        self.post_event("mb1_start")
        self.assertEqual(1, self.machine.multiballs["mb1"].balls_added_live)

        # another ball should be ejected to pf
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertBallsInPlay(2)

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # game should not end
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # it should be readded because of shoot again
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertBallsInPlay(2)

        # two balls drain
        self.drain_one_ball()
        self.drain_one_ball()
        self.assertEqual(0, self._events['multiball_mb1_ended'])

        # they should be readded because of shoot again
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertEventNotCalled("multiball_mb1_ball_lost")
        self.assertBallsInPlay(2)

        # shoot again ends
        self.advance_time_and_run(10)

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)
        self.assertEventCalled("multiball_mb1_ball_lost", 1)
        self.assertBallsInPlay(1)

        # mb ends
        self.assertEqual(1, self._events['multiball_mb1_ended'])

        # the other ball also drains
        self.drain_one_ball()

        # game should end
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

    def testRestartMultiball(self):
        self.mock_event("multiball_mb1_ended")

        # prepare game
        self.fill_troughs()
        self.assertFalse(self.machine.multiballs["mb1"].enabled)

        # start game
        self.start_game()
        self.post_event("mb1_enable")

        # multiball should be enabled
        self.assertTrue(self.machine.multiballs["mb1"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # mb started
        self.post_event("mb1_start")
        self.assertEqual(1, self.machine.multiballs["mb1"].balls_added_live)

        # another ball should be ejected to pf
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.playfield.balls)

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # game should not end
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # it should be readded because of shoot again
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertEqual(2, self.machine.playfield.available_balls)

        # mb cannot start again/nothing happens
        self.post_event("mb1_start")
        self.assertEqual(1, self.machine.multiballs["mb1"].balls_added_live)
        self.assertEqual(2, self.machine.playfield.available_balls)

        # shoot again ends
        self.advance_time_and_run(10)

        # mb cannot start again because balls are still in play
        self.post_event("mb1_start")
        self.assertEqual(1, self.machine.multiballs["mb1"].balls_added_live)
        self.assertEqual(2, self.machine.playfield.available_balls)
        self.assertEqual(0, self._events['multiball_mb1_ended'])

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # mb ends
        self.assertEqual(0, self.machine.multiballs["mb1"].balls_added_live)
        self.assertEqual(1, self._events['multiball_mb1_ended'])

        # restart mb
        self.post_event("mb1_start")
        self.advance_time_and_run(1)
        self.assertEqual(1, self.machine.multiballs["mb1"].balls_added_live)
        self.assertEqual(2, self.machine.playfield.available_balls)
        self.assertEqual(1, self.machine.playfield.balls)

        self.advance_time_and_run(40)
        self.assertEqual(2, self.machine.playfield.balls)

        # two balls drains
        self.drain_one_ball()
        self.drain_one_ball()

        # game should end
        self.advance_time_and_run(1)
        self.assertEqual(2, self._events['multiball_mb1_ended'])
        self.assertEqual(None, self.machine.game)

    def testUnlimitedShootAgain(self):
        self.mock_event("multiball_mb2_ended")

        # prepare game
        self.fill_troughs()

        # start game
        self.start_game()

        self.post_event("mb2_enable")

        # multiball should be enabled
        self.assertTrue(self.machine.multiballs["mb2"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # mb started
        self.post_event("mb2_start")
        self.assertEqual(2, self.machine.multiballs["mb2"].balls_added_live)

        # another two balls should be ejected to pf
        self.advance_time_and_run(10)
        self.assertEqual(3, self.machine.playfield.balls)

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # game should not end
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(2, self.machine.playfield.balls)

        # it should be readded because of shoot again
        self.advance_time_and_run(10)
        self.assertEqual(3, self.machine.playfield.balls)

        # two balls drain
        self.drain_one_ball()
        self.drain_one_ball()
        self.advance_time_and_run(1)
        self.assertEqual(0, self._events['multiball_mb2_ended'])

        # they should be readded because of shoot again
        self.advance_time_and_run(10)
        self.assertEqual(3, self.machine.playfield.balls)

        # shoot again forever
        self.advance_time_and_run(100)

        # three balls drain
        self.drain_one_ball()
        self.drain_one_ball()
        self.drain_one_ball()
        self.advance_time_and_run(1)
        self.assertEqual(0, self.machine.playfield.balls)

        # they should be readded because of shoot again
        self.advance_time_and_run(20)
        self.assertEqual(3, self.machine.playfield.balls)

        # end mb
        self.post_event("mb2_stop")

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # mb does not end yet
        self.assertEqual(0, self._events['multiball_mb2_ended'])

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # mb ends
        self.assertEqual(1, self._events['multiball_mb2_ended'])

        # the other ball also drains
        self.drain_one_ball()

        # game should end
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

    def testSimultaneousMultiballs(self):
        self.mock_event("multiball_mb2_ended")
        self.mock_event("multiball_mb3_ended")

        # prepare game
        self.fill_troughs()

        # start game
        self.start_game()

        self.post_event("mb2_enable")
        self.post_event("mb3_enable")

        # multiballs should be enabled
        self.assertTrue(self.machine.multiballs["mb2"].enabled)
        self.assertTrue(self.machine.multiballs["mb3"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # mb started
        self.post_event("mb2_start")
        self.assertEqual(2, self.machine.multiballs["mb2"].balls_added_live)

        # another two balls should be ejected to pf
        self.advance_time_and_run(10)
        self.assertEqual(3, self.machine.playfield.balls)

        # start mb3
        self.post_event("mb3_start")
        self.assertEqual(1, self.machine.multiballs["mb3"].balls_added_live)

        # another ball should appear
        self.advance_time_and_run(10)
        self.assertEqual(4, self.machine.playfield.balls)

        self.assertTrue(self.machine.multiballs["mb2"].shoot_again)
        self.assertFalse(self.machine.multiballs["mb3"].shoot_again)

        self.post_event("mb2_stop")
        self.assertFalse(self.machine.multiballs["mb2"].shoot_again)

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # game should not end
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(3, self.machine.playfield.balls)
        self.assertEqual(0, self._events['multiball_mb2_ended'])
        self.assertEqual(0, self._events['multiball_mb3_ended'])

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertEqual(0, self._events['multiball_mb2_ended'])
        self.assertEqual(0, self._events['multiball_mb3_ended'])

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, self._events['multiball_mb2_ended'])
        self.assertEqual(1, self._events['multiball_mb3_ended'])

        # last ball drains
        self.drain_one_ball()

        # game should end
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

    def testMultiballInMode(self):
        self.mock_event("multiball_mb4_ended")

        # prepare game
        self.fill_troughs()

        # start game
        self.start_game()

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # mode not loaded. mb4 should not enable or start
        self.post_event("mb4_enable")
        self.post_event("mb4_start")

        self.advance_time_and_run(4)
        self.assertEqual(1, self.machine.playfield.available_balls)

        # start mode
        self.post_event("start_mode1")

        # mode loaded. mb4 should enable and start
        self.post_event("mb4_enable")
        self.post_event("mb4_start")
        self.assertTrue(self.machine.multiballs["mb4"].enabled)

        self.advance_time_and_run(4)
        self.assertEqual(2, self.machine.playfield.available_balls)

        # another ball should be ejected to pf
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.playfield.balls)

        self.drain_one_ball()
        self.advance_time_and_run(1)

        # it should come back
        self.assertEqual(2, self.machine.playfield.available_balls)
        self.assertEqual(0, self._events['multiball_mb4_ended'])

        # stop mode
        self.post_event("stop_mode1")

        # mode end should stop mp
        self.assertFalse(self.machine.multiballs["mb4"].shoot_again)
        self.assertFalse(self.machine.multiballs["mb4"].enabled)
        self.assertEqual(0, self._events['multiball_mb4_ended'])

        # next drain should end mb
        self.drain_one_ball()
        self.advance_time_and_run(1)
        self.assertEqual(1, self.machine.playfield.available_balls)
        self.assertEqual(1, self._events['multiball_mb4_ended'])

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # game should end
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

    def testMultiballInModeSimple(self):
        self.mock_event("multiball_mb5_ended")

        # prepare game
        self.fill_troughs()

        # start game
        self.start_game()

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        self.advance_time_and_run(4)
        self.assertEqual(1, self.machine.playfield.available_balls)

        # start mode
        self.post_event("start_mode2")

        # mode loaded. mb5 should enabled but not started
        self.assertTrue(self.machine.multiballs["mb5"].enabled)
        self.assertEqual(0, self.machine.multiballs["mb5"].balls_added_live)

        # start it
        self.post_event("mb5_start")
        self.assertEqual(1, self.machine.multiballs["mb5"].balls_added_live)

        self.advance_time_and_run(4)
        self.assertEqual(2, self.machine.playfield.available_balls)

        # drain a ball
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # it should come back
        self.assertEqual(2, self.machine.playfield.available_balls)
        self.assertEqual(0, self._events['multiball_mb5_ended'])

        # stop mode
        self.post_event("stop_mode2")

        # mode end should stop mb
        self.assertFalse(self.machine.multiballs["mb5"].shoot_again)
        self.assertFalse(self.machine.multiballs["mb5"].enabled)
        self.assertEqual(0, self._events['multiball_mb5_ended'])

        # next drain should end mb
        self.drain_one_ball()
        self.advance_time_and_run(1)
        self.assertEqual(1, self.machine.playfield.available_balls)
        self.assertEqual(1, self._events['multiball_mb5_ended'])

        # ball drains
        self.drain_one_ball()
        self.advance_time_and_run(1)

        # game should end
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

    def testMultiballWithLock(self):
        # prepare game
        self.fill_troughs()
        self.assertFalse(self.machine.multiballs["mb6"].enabled)

        # start game
        self.start_game()

        # start mode
        self.post_event("start_mode1")

        # multiball should be enabled
        self.assertTrue(self.machine.multiballs["mb6"].enabled)
        # lock should be enabled
        self.assertTrue(self.machine.multiball_locks["lock_mb6"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # lock one ball and another one should go to pf
        self.hit_switch_and_run("s_lock1", 10)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, self.machine.game.player_list[0]["lock_mb6_locked_balls"])

        # start mb
        self.post_event("mb6_start")
        self.assertEqual(2, self.machine.multiballs["mb6"].balls_added_live)

        # three balls on pf
        self.advance_time_and_run(10)
        self.assertEqual(3, self.machine.playfield.balls)
        self.assertEqual(0, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(0, self.machine.game.player_list[0]["lock_mb6_locked_balls"])

        # game ends (because of slam tilt)
        self.machine.game.end_ball()
        self.advance_time_and_run()

        # this should not crash
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_trough"])
        self.advance_time_and_run()

    def test_total_ball_count(self):
        # prepare game
        self.fill_troughs()
        self.assertFalse(self.machine.multiballs["mb10"].enabled)
        self.assertFalse(self.machine.multiballs["mb11"].enabled)

        # start game
        self.start_game()

        # start mode
        self.post_event("start_mode1")

        # multiball should be enabled
        self.assertTrue(self.machine.multiballs["mb10"].enabled)
        self.assertTrue(self.machine.multiballs["mb11"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # start mb10
        self.post_event("mb10_start")
        self.assertEqual(2, self.machine.multiballs["mb10"].balls_added_live)
        self.assertEqual(3, self.machine.multiballs["mb10"].balls_live_target)
        self.assertTrue(self.machine.multiballs["mb10"].shoot_again)

        # three balls on pf
        self.advance_time_and_run(10)
        self.assertEqual(3, self.machine.playfield.balls)

        # drain one. should come back
        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.assertEqual(3, self.machine.playfield.balls)
        self.assertEqual(2, self.machine.multiballs["mb10"].balls_added_live)
        self.assertEqual(3, self.machine.multiballs["mb10"].balls_live_target)
        self.assertTrue(self.machine.multiballs["mb10"].shoot_again)

        # no more shoot again
        self.advance_time_and_run(5)
        self.assertFalse(self.machine.multiballs["mb10"].shoot_again)

        # start mb11
        self.post_event("mb11_start")
        self.assertEqual(0, self.machine.multiballs["mb11"].balls_added_live)
        self.assertEqual(2, self.machine.multiballs["mb11"].balls_live_target)
        self.advance_time_and_run(5)
        self.assertEqual(3, self.machine.playfield.balls)
        self.assertEqual(3, self.machine.game.balls_in_play)
        self.assertTrue(self.machine.multiballs["mb11"].shoot_again)

        # drain one. should not come back
        self.drain_one_ball()
        self.advance_time_and_run(4)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertEqual(2, self.machine.game.balls_in_play)
        self.assertTrue(self.machine.multiballs["mb11"].shoot_again)

        # but the second one should come back
        self.drain_one_ball()
        self.advance_time_and_run(4)
        self.assertEqual(2, self.machine.playfield.balls)
        self.assertEqual(2, self.machine.game.balls_in_play)
        self.assertTrue(self.machine.multiballs["mb11"].shoot_again)

        # shoot again ends
        self.advance_time_and_run(10)
        self.assertFalse(self.machine.multiballs["mb10"].shoot_again)
        self.assertFalse(self.machine.multiballs["mb11"].shoot_again)
        self.assertEqual(3, self.machine.multiballs["mb10"].balls_live_target)
        self.assertEqual(2, self.machine.multiballs["mb11"].balls_live_target)
        self.assertEqual(2, self.machine.game.balls_in_play)

        # drain one balls
        self.drain_one_ball()
        self.advance_time_and_run()
        self.assertEqual(1, self.machine.game.balls_in_play)

        # both mbs should end
        self.assertEqual(0, self.machine.multiballs["mb10"].balls_live_target)
        self.assertEqual(0, self.machine.multiballs["mb11"].balls_live_target)

    def test_total_ball_count_with_lock(self):
        # prepare game
        self.fill_troughs()
        self.assertFalse(self.machine.multiballs["mb10"].enabled)
        self.assertFalse(self.machine.multiballs["mb11"].enabled)

        # start game
        self.start_game()

        # start mode
        self.post_event("start_mode1")

        # multiball should be enabled
        self.assertTrue(self.machine.multiballs["mb10"].enabled)
        self.assertTrue(self.machine.multiballs["mb11"].enabled)
        # lock should be enabled
        self.assertTrue(self.machine.multiball_locks["lock_mb6"].enabled)

        # takes roughly 4s to get ball confirmed
        self.advance_time_and_run(4)
        self.assertNotEqual(None, self.machine.game)
        self.assertEqual(1, self.machine.playfield.balls)

        # lock one ball and another one should go to pf
        self.hit_switch_and_run("s_lock1", 10)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(1, self.machine.playfield.balls)

        # start mb10
        self.post_event("mb10_start")
        self.assertEqual(2, self.machine.multiballs["mb10"].balls_added_live)
        self.assertEqual(3, self.machine.multiballs["mb10"].balls_live_target)

        # three balls on pf and one in lock
        self.advance_time_and_run(10)
        self.assertEqual(3, self.machine.playfield.balls)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)

        # start mb12. eject lock
        self.post_event("mb12_start")
        self.assertEqual(1, self.machine.multiballs["mb12"].balls_added_live)
        self.assertEqual(4, self.machine.multiballs["mb12"].balls_live_target)
        self.advance_time_and_run(5)
        self.assertEqual(4, self.machine.playfield.balls)
        self.assertEqual(4, self.machine.game.balls_in_play)
        self.assertTrue(self.machine.multiballs["mb12"].shoot_again)

    def testAddABall(self):
        self.mock_event("multiball_mb_add_a_ball_ended")

        # prepare game
        self.fill_troughs()

        # start game
        self.start_game()

        self.post_event("start_or_add")
        self.advance_time_and_run(10)
        self.assertBallsOnPlayfield(2)

        self.post_event("start_or_add")
        self.advance_time_and_run(5)
        self.assertBallsOnPlayfield(3)

        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.assertBallsOnPlayfield(2)
        self.assertEventNotCalled("multiball_mb_add_a_ball_ended")

        self.post_event("add_ball")
        self.advance_time_and_run(5)
        self.assertBallsOnPlayfield(3)

        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.assertBallsOnPlayfield(2)
        self.assertEventNotCalled("multiball_mb_add_a_ball_ended")

        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.assertBallsOnPlayfield(1)
        self.assertEventCalled("multiball_mb_add_a_ball_ended")

        self.post_event("add_ball")
        self.advance_time_and_run(5)
        self.assertBallsOnPlayfield(1)

    def testMultiballLockFullMultiplayer(self):
        self.machine.config['game']['balls_per_game'] = self.machine.placeholder_manager.build_int_template(3)
        self.mock_event("multiball_lock_lock_mb6_full")
        self.fill_troughs()
        self.start_two_player_game()
        self.assertPlayerNumber(1)
        self.assertBallNumber(1)

        self.post_event("start_mode1")

        # lock ball
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(5)

        # machine should request a new ball and the lock keeps one
        self.assertEqual(1, self.machine.game.player_list[0]["lock_mb6_locked_balls"])
        self.assertEqual(0, self.machine.game.player_list[1]["lock_mb6_locked_balls"])
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(4, self.machine.ball_devices["bd_trough"].balls)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)

        # drain ball. player 2 should be up
        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.post_event("start_mode1")
        self.assertPlayerNumber(2)
        self.assertBallNumber(1)

        # also lock a ball
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(5)

        # lock should not keep the ball but count it for the player
        self.assertEqual(1, self.machine.game.player_list[0]["lock_mb6_locked_balls"])
        self.assertEqual(1, self.machine.game.player_list[1]["lock_mb6_locked_balls"])
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(4, self.machine.ball_devices["bd_trough"].balls)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)
        self.assertEventNotCalled("multiball_lock_lock_mb6_full")

        # lock another ball. lock should keep it
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(5)
        self.assertEqual(1, self.machine.game.player_list[0]["lock_mb6_locked_balls"])
        self.assertEqual(2, self.machine.game.player_list[1]["lock_mb6_locked_balls"])
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(3, self.machine.ball_devices["bd_trough"].balls)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)
        self.assertEventCalled("multiball_lock_lock_mb6_full")

        # drain ball. lock should release a ball because player1 need to be able to complete it
        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.assertEqual(1, self.machine.game.player_list[0]["lock_mb6_locked_balls"])
        self.assertEqual(2, self.machine.game.player_list[1]["lock_mb6_locked_balls"])
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(4, self.machine.ball_devices["bd_trough"].balls)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(0)

        # ball from lock drains
        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.assertPlayerNumber(1)
        self.assertBallNumber(2)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(4, self.machine.ball_devices["bd_trough"].balls)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)

        self.post_event("start_mode1")

        # lock another ball. lock should keep it
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(5)
        self.assertEqual(2, self.machine.game.player_list[0]["lock_mb6_locked_balls"])
        self.assertEqual(2, self.machine.game.player_list[1]["lock_mb6_locked_balls"])
        self.assertEqual(2, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(3, self.machine.ball_devices["bd_trough"].balls)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)

        # start MB
        self.post_event("mb6_start")
        self.advance_time_and_run(5)
        self.assertBallsOnPlayfield(3)
        self.assertBallsInPlay(3)
        self.assertEqual(0, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(3, self.machine.ball_devices["bd_trough"].balls)

        # drain ball
        self.drain_one_ball()
        self.advance_time_and_run()
        self.assertPlayerNumber(1)
        self.assertBallNumber(2)
        self.assertBallsOnPlayfield(2)
        self.assertBallsInPlay(2)

        # drain ball
        self.drain_one_ball()
        self.advance_time_and_run()
        self.assertPlayerNumber(1)
        self.assertBallNumber(2)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)

        # drain ball
        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.post_event("start_mode1")
        self.assertPlayerNumber(2)
        self.assertBallNumber(2)
        self.assertEqual(0, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(5, self.machine.ball_devices["bd_trough"].balls)
        self.assertEqual(0, self.machine.game.player_list[0]["lock_mb6_locked_balls"])
        self.assertEqual(2, self.machine.game.player_list[1]["lock_mb6_locked_balls"])

        # start mb without balls in lock
        self.post_event("mb6_start")
        self.advance_time_and_run(15)
        self.assertBallsOnPlayfield(3)
        self.assertBallsInPlay(3)
        self.assertEqual(0, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(3, self.machine.ball_devices["bd_trough"].balls)

        self.assertEqual(0, self.machine.game.player_list[0]["lock_mb6_locked_balls"])
        self.assertEqual(0, self.machine.game.player_list[1]["lock_mb6_locked_balls"])

        # drain ball
        self.drain_one_ball()
        self.advance_time_and_run()
        self.assertPlayerNumber(2)
        self.assertBallNumber(2)
        self.assertBallsOnPlayfield(2)
        self.assertBallsInPlay(2)

        # drain ball
        self.drain_one_ball()
        self.advance_time_and_run()
        self.assertPlayerNumber(2)
        self.assertBallNumber(2)
        self.assertBallsOnPlayfield(1)
        self.assertBallsInPlay(1)

        # drain last ball
        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.post_event("start_mode1")
        self.assertPlayerNumber(1)
        self.assertBallNumber(3)
        self.assertEqual(0, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(5, self.machine.ball_devices["bd_trough"].balls)
        self.assertBallsInPlay(1)

        # lock ball
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(5)
        self.assertEqual(1, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(4, self.machine.ball_devices["bd_trough"].balls)
        self.assertBallsInPlay(1)

        # drain again
        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.assertPlayerNumber(2)
        self.assertBallNumber(3)

        self.assertEqual(1, self.machine.game.player_list[0]["lock_mb6_locked_balls"])
        self.assertEqual(0, self.machine.game.player_list[1]["lock_mb6_locked_balls"])

        # drain again. game should end
        self.drain_one_ball()

        # lock should eject all balls
        self.advance_time_and_run(5)
        self.assertGameIsNotRunning()
        self.assertEqual(0, self.machine.ball_devices["bd_lock"].balls)
        self.assertEqual(5, self.machine.ball_devices["bd_trough"].balls)
        self.assertBallsOnPlayfield(1)

        # game should not start yet
        self.assertGameIsNotRunning()
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertGameIsNotRunning()

        # ball from lock drain
        self.drain_one_ball()
        self.advance_time_and_run()

        # start new game
        self.start_game()
        self.post_event("start_mode1")

        self.drain_one_ball()
        self.advance_time_and_run()

    def testModeWithMultiballAutostart(self):
        # prepare game
        self.fill_troughs()

        # start game
        self.start_game()
        self.post_event("start_mode3")
        self.advance_time_and_run(1)

        # multiball should be enabled now but not started
        self.assertTrue(self.machine.multiballs["mb_autostart"].enabled)
        self.assertEqual(1, self.machine.multiballs["mb_autostart"].balls_added_live)

    def testMultiballWhichStartsAfterLock(self):
        self.mock_event("multiball_mb_autostart_ended")
        self.mock_event("multiball_mb_autostart_ball_lost")

        # prepare game
        self.fill_troughs()
        self.assertFalse(self.machine.multiballs["mb4_autostart"].enabled)

        # start game
        self.start_game()

        # start mode
        self.post_event("start_mode4")
        self.advance_time_and_run(5)

        self.assertAvailableBallsOnPlayfield(1)
        self.assertEqual(5, self.machine.ball_devices["bd_trough"].available_balls)

        # multiball should be enabled now but not started
        self.assertTrue(self.machine.multiballs["mb4_autostart"].enabled)
        self.assertEqual(0, self.machine.multiballs["mb4_autostart"].balls_added_live)

        # lock a ball
        self.machine.default_platform.add_ball_to_device(self.machine.ball_devices["bd_lock"])
        self.advance_time_and_run(1)

        # mb should start
        self.assertTrue(self.machine.multiballs["mb4_autostart"].enabled)
        self.assertEqual(1, self.machine.multiballs["mb4_autostart"].balls_added_live)
        self.assertBallsInPlay(2)
        self.assertAvailableBallsOnPlayfield(2)

        # lock should eject
        self.assertEqual(4, self.machine.ball_devices["bd_trough"].available_balls)
        self.assertEqual(0, self.machine.ball_devices["bd_lock"].available_balls)

        # both balls drain
        self.drain_all_balls()

        # game should end
        self.advance_time_and_run(1)
        self.assertEqual(None, self.machine.game)

    def testMultiballStateInPlaceholder(self):
        self.fill_troughs()
        self.start_game()

        self.post_event("start_default")
        self.mock_event("should_post_when_enabled")
        self.mock_event("should_post_when_disabled")
        self.mock_event("should_not_post_when_enabled")
        self.mock_event("should_not_post_when_disabled")
        mb = self.machine.multiballs["mb1"]

        self.assertFalse(mb.enabled)
        self.post_event("test_event_when_disabled")
        self.assertEventCalled("should_post_when_disabled")
        self.assertEventNotCalled("should_not_post_when_disabled")

        mb.enable()
        self.assertTrue(mb.enabled)
        self.post_event("test_event_when_enabled")
        self.assertEventCalled("should_post_when_enabled")
        self.assertEventNotCalled("should_not_post_when_enabled")

    def testShootAgainPlaceholder(self):
        self.fill_troughs()
        self.start_game()
        self.assertAvailableBallsOnPlayfield(1)

        # start mb with no shoot again set in machine var
        self.mock_event("multiball_mb_placeholder_shoot_again_ended")
        self.mock_event("multiball_mb_placeholder_ended")
        self.post_event("mb_placeholder_start")
        self.advance_time_and_run(5)
        self.assertAvailableBallsOnPlayfield(2)
        # shoot again should end instantly
        self.assertEventCalled("multiball_mb_placeholder_shoot_again_ended")

        # drain one ball
        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.assertAvailableBallsOnPlayfield(1)
        # mb should end
        self.assertEventCalled("multiball_mb_placeholder_ended")

        # set shoot again time
        self.machine.variables.set_machine_var("shoot_again_sec", 30)
        # start mb again
        self.mock_event("multiball_mb_placeholder_shoot_again_ended")
        self.mock_event("multiball_mb_placeholder_ended")
        self.post_event("mb_placeholder_start")
        self.advance_time_and_run(5)
        self.assertAvailableBallsOnPlayfield(2)
        # shoot again should not end instantly
        self.assertEventNotCalled("multiball_mb_placeholder_shoot_again_ended")

        # drain one ball
        self.drain_one_ball()
        self.advance_time_and_run(5)
        # shoot again should bring it back
        self.assertAvailableBallsOnPlayfield(2)
        self.assertEventNotCalled("multiball_mb_placeholder_ended")

        # wait 30s for shoot again to end
        self.advance_time_and_run(30)
        self.assertEventCalled("multiball_mb_placeholder_shoot_again_ended")

        # drain one ball
        self.drain_one_ball()
        self.advance_time_and_run(5)
        self.assertAvailableBallsOnPlayfield(1)
        # mb should end
        self.assertEventCalled("multiball_mb_placeholder_ended")



"""A shot in MPF."""
from mpf.core.device_monitor import DeviceMonitor

from mpf.core.enable_disable_mixin import EnableDisableMixin

import mpf.core.delays
from mpf.core.events import event_handler
from mpf.core.mode import Mode
from mpf.core.mode_device import ModeDevice
from mpf.core.player import Player


@DeviceMonitor("state", "state_name")
class Shot(EnableDisableMixin, ModeDevice):

    """A device which represents a generic shot."""

    config_section = 'shots'
    collection = 'shots'
    class_label = 'shot'

    monitor_enabled = False
    """Class attribute which specifies whether any monitors have been registered
    to track shots.
    """

    __slots__ = ["delay", "active_sequences", "active_delays", "running_show", "_handlers"]

    def __init__(self, machine, name):
        """Initialise shot."""
        # If this device is setup in a machine-wide config, make sure it has
        # a default enable event.
        super(Shot, self).__init__(machine, name)

        self.delay = mpf.core.delays.DelayManager(self.machine)

        self.active_sequences = list()
        """List of tuples: (id, current_position_index, next_switch)"""
        self.active_delays = set()
        self.running_show = None
        self._handlers = []

    async def _initialize(self) -> None:
        """Register playfield active handlers."""
        await super()._initialize()

        for switch in self.config['switches'] + list(self.config['delay_switch'].keys()):
            # mark the playfield active no matter what
            switch.add_handler(self._mark_active)

    def _mark_active(self, **kwargs):
        """Mark playfield active."""
        del kwargs
        self.config['playfield'].mark_playfield_active_from_device_action()

    def device_loaded_in_mode(self, mode: Mode, player: Player):
        """Add device to a mode that was already started.

        Automatically enables the shot and calls the the method
        that's usually called when a player's turn starts since that was missed
        since the mode started after that.
        """
        super().device_loaded_in_mode(mode, player)
        self._update_show()

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None):
        """Validate and parse shot config."""
        config = super().validate_and_parse_config(config, is_mode_config, debug_prefix)
        for switch in config['switch']:
            if switch not in config['switches']:
                config['switches'].append(switch)

        for switch in config['switches'] + list(config['delay_switch'].keys()):
            if '{}_active'.format(config['playfield'].name) in switch.tags:
                self.raise_config_error(
                    "Shot '{}' uses switch '{}' which has a "
                    "'{}_active' tag. This is handled internally by the device. Remove the "
                    "redundant '{}_active' tag from that switch.".format(
                        self.name, switch.name, config['playfield'].name,
                        config['playfield'].name), 1)

        return config

    def _register_switch_handlers(self):
        self._handlers = []
        for switch in self.config['switches']:
            self._handlers.append(self.machine.events.add_handler("{}_active".format(switch.name),
                                                                  self.event_hit, priority=self.mode.priority,
                                                                  blocking_facility="shot"))

        for switch in list(self.config['delay_switch'].keys()):
            self._handlers.append(self.machine.events.add_handler("{}_active".format(switch.name),
                                                                  self._delay_switch_hit,
                                                                  switch_name=switch.name,
                                                                  priority=self.mode.priority,
                                                                  blocking_facility="shot"))

    def _remove_switch_handlers(self):
        self.delay.clear()
        self.active_delays = set()

        self.machine.events.remove_handlers_by_keys(self._handlers)
        self._handlers = []

    @event_handler(6)
    def event_advance(self, force=False, **kwargs):
        """Handle advance control event."""
        del kwargs
        self.advance(force)

    def advance(self, force=False) -> bool:
        """Advance a shot profile forward.

        If this profile is at the last step and configured to loop, it will
        roll over to the first step. If this profile is at the last step and not
        configured to loop, this method has no effect.
        """
        if not self.enabled and not force:
            return False

        if not self.player:
            # no player no state
            return False

        profile_name = self.config['profile'].name
        state = self._get_state()

        self.debug_log("Advancing 1 step. Profile: %s, "
                       "Current State: %s", profile_name, state)

        if state + 1 >= len(self.config['profile'].config['states']):

            if self.config['profile'].config['loop']:
                self._set_state(0)

            else:
                return False
        else:
            self.debug_log("Advancing shot by one step.")
            self._set_state(state + 1)

        self._update_show()
        return True

    def _stop_show(self):
        if not self.running_show:
            return
        self.running_show.stop()
        self.running_show = None

    @property
    def can_rotate(self):
        """Return if the shot can be rotated according to its profile."""
        state = self.state_name
        return state not in self.profile.config['state_names_to_not_rotate']

    @property
    def state_name(self):
        """Return current state name."""
        if not self.player:
            # no player no state
            return "None"
        return self.config['profile'].config['states'][self._get_state()]['name']

    @property
    def state(self):
        """Return current state index."""
        return self._get_state()

    @property
    def profile_name(self):
        """Return profile name."""
        return self.config['profile'].name

    @property
    def profile(self):
        """Return profile."""
        return self.config['profile']

    def _get_state(self):
        if not self.player:
            return 0
        return self.player["shot_{}".format(self.name)]

    def _set_state(self, state):
        old = self.player["shot_{}".format(self.name)]
        try:
            old_name = self.state_name
        except IndexError:
            # In this case, the shot profile was changed and the old state index
            # doesn't exist in the new profile. That's okay, but we can't include
            # the old state name in our event.
            old_name = "unknown"
        self.player["shot_{}".format(self.name)] = state
        self.notify_virtual_change("state", old, state)
        self.notify_virtual_change("state_name", old_name, self.state_name)

    def _get_profile_settings(self):
        state = self._get_state()
        return self.profile.config['states'][state]

    def _update_show(self):
        if not self.enabled and not self.profile.config['show_when_disabled']:
            self._stop_show()
            return

        state = self._get_state()
        state_settings = self.profile.config['states'][state]

        if state_settings['show']:  # there's a show specified this state
            self._play_show(settings=state_settings)

        elif self.profile.config['show']:
            # no show for this state, but we have a profile root show
            self._play_show(settings=state_settings, start_step=state + 1)

        # if neither if/elif above happens, it means the current step has no
        # show but the previous step had one. We stop the previous show if there is one
        elif self.running_show:
            self._stop_show()

    def _play_show(self, settings, start_step=None):
        manual_advance = settings['manual_advance']
        if settings['show']:
            show_name = settings['show']
            if settings['manual_advance'] is None:
                manual_advance = False
        else:
            show_name = self.profile.config['show']
            if settings['manual_advance'] is None:
                manual_advance = True

        if settings['show_tokens'] and self.config['show_tokens']:
            show_tokens = dict(settings['show_tokens'])
            show_tokens.update(self.config['show_tokens'])
        elif settings['show_tokens']:
            show_tokens = settings['show_tokens']
        elif self.config['show_tokens']:
            show_tokens = self.config['show_tokens']
        else:
            show_tokens = {}

        if show_tokens:
            show_tokens = {k: v.evaluate({})
                           for k, v in show_tokens.items()}

        priority = settings['priority'] + self.mode.priority
        if not start_step:
            start_step = settings['start_step']

        self.debug_log("Playing show: %s. %s", show_name, settings)

        show_config = self.machine.show_controller.create_show_config(
            show_name, priority=priority, speed=settings.get("speed"),
            loops=settings.get("loops", -1), sync_ms=settings.get("sync_ms"), manual_advance=manual_advance,
            show_tokens=show_tokens, events_when_played=settings.get("events_when_played"),
            events_when_stopped=settings.get("events_when_stopped"),
            events_when_looped=settings.get("events_when_looped"),
            events_when_paused=settings.get("events_when_paused"),
            events_when_resumed=settings.get("events_when_resumed"),
            events_when_advanced=settings.get("events_when_advanced"),
            events_when_stepped_back=settings.get("events_when_stepped_back"),
            events_when_updated=settings.get("events_when_updated"),
            events_when_completed=settings.get("events_when_completed"))
        self.running_show = self.machine.show_controller.replace_or_advance_show(self.running_show, show_config,
                                                                                 start_step)

    def device_removed_from_mode(self, mode):
        """Remove this shot device.

        Destroys it and removes it from the shots collection.
        """
        super().device_removed_from_mode(mode)
        self._remove_switch_handlers()
        if self.running_show:
            self.running_show.stop()
            self.running_show = None

    @event_handler(5)
    def event_hit(self, **kwargs):
        """Handle hit control event."""
        success = self.hit()
        if not success:
            return None

        if self.profile.config['block']:
            min_priority = kwargs.get("_min_priority", {"all": 0})
            min_shots = min_priority.get("shot", 0)
            min_priority["shot"] = self.mode.priority if self.mode.priority > min_shots else min_shots
            return {"_min_priority": min_priority}

        return None

    def hit(self) -> bool:
        """Advance the currently-active shot profile.

        Note that the shot must be enabled in order for this hit to be
        processed.

        Returns true if the shot was enabled or false if the hit has been ignored.
        """
        if not self.enabled or not self.player:
            return False

        # Stop if there is an active delay but no sequence
        if self.active_delays:
            return False

        profile_settings = self._get_profile_settings()

        if not profile_settings:
            return False

        state = profile_settings['name']

        self.debug_log("Hit! Profile: %s, State: %s",
                       self.profile_name, state)

        if self.profile.config['advance_on_hit']:
            self.debug_log("Advancing shot because advance_on_hit is True.")
            advancing = self.advance()
        else:
            self.debug_log('Not advancing shot')
            advancing = False

        self._notify_monitors(self.config['profile'].name, state)

        self.machine.events.post('{}_hit'.format(self.name),
                                 profile=self.profile_name, state=state, advancing=advancing)
        '''event: (name)_hit
        desc: The shot called (name) was just hit.

        Note that there are four events posted when a shot is hit, each
        with variants of the shot name, profile, and current state,
        allowing you to key in on the specific granularity you need.

        args:
        profile: The name of the profile that was active when hit.
        state: The name of the state the profile was in when it was hit'''

        self.machine.events.post('{}_{}_hit'.format(self.name, self.profile_name),
                                 profile=self.profile_name, state=state, advancing=advancing)
        '''event: (name)_(profile)_hit
        desc: The shot called (name) was just hit with the profile (profile)
        active.

        Note that there are four events posted when a shot is hit, each
        with variants of the shot name, profile, and current state,
        allowing you to key in on the specific granularity you need.

        Also remember that shots can have more than one active profile at a
        time (typically each associated with a mode), so a single hit to this
        shot might result in this event being posted multiple times with
        different (profile) values.

        args:
        profile: The name of the profile that was active when hit.
        state: The name of the state the profile was in when it was hit'''

        self.machine.events.post('{}_{}_{}_hit'.format(self.name, self.profile_name, state),
                                 profile=self.profile_name, state=state, advancing=advancing)
        '''event: (name)_(profile)_(state)_hit
        desc: The shot called (name) was just hit with the profile (profile)
        active in the state (state).

        Note that there are four events posted when a shot is hit, each
        with variants of the shot name, profile, and current state,
        allowing you to key in on the specific granularity you need.

        Also remember that shots can have more than one active profile at a
        time (typically each associated with a mode), so a single hit to this
        shot might result in this event being posted multiple times with
        different (profile) and (state) values.

        args:
        profile: The name of the profile that was active when hit.
        state: The name of the state the profile was in when it was hit'''

        self.machine.events.post('{}_{}_hit'.format(self.name, state),
                                 profile=self.profile_name, state=state, advancing=advancing)
        '''event: (name)_(state)_hit
        desc: The shot called (name) was just hit while in the profile (state).

        Note that there are four events posted when a shot is hit, each
        with variants of the shot name, profile, and current state,
        allowing you to key in on the specific granularity you need.

        Also remember that shots can have more than one active profile at a
        time (typically each associated with a mode), so a single hit to this
        shot might result in this event being posted multiple times with
        different (profile) and (state) values.

        args:
        profile: The name of the profile that was active when hit.
        state: The name of the state the profile was in when it was hit'''

        return True

    def _notify_monitors(self, profile, state):
        if Shot.monitor_enabled and "shots" in self.machine.monitors:
            for callback in self.machine.monitors['shots']:
                callback(name=self.name, profile=profile, state=state)

    @event_handler(4)
    def _delay_switch_hit(self, switch_name, **kwargs):
        del kwargs
        if not self.enabled:
            return

        self.delay.reset(name=switch_name + '_delay_timer',
                         ms=self.config['delay_switch']
                                       [self.machine.switches[switch_name]],
                         callback=self._release_delay,
                         switch=switch_name)

        self.active_delays.add(switch_name)

    def _release_delay(self, switch):
        self.active_delays.remove(switch)

    def jump(self, state, force=True, force_show=False):
        """Jump to a certain state in the active shot profile.

        Args:
        ----
            state: int of the state number you want to jump to. Note that states
                are zero-based, so the first state is 0.
            force: if true, will jump even if the shot is disabled
            force_show: if true, will update the profile show even if the jumped
                state index is the same as before the jump

        """
        self.debug_log("Received jump request. State: %s, Force: %s", state, force)

        if not self.enabled and not force:
            self.debug_log("Profile is disabled and force is False. Not jumping")
            return

        if not self.player:
            # no player no state
            return

        current_state = self._get_state()
        if state == current_state and not force_show:
            self.debug_log("Shot is already in the jump destination state")
            return

        self.debug_log("Jumping to profile state '%s'", state)

        self._set_state(state)

        self._update_show()

    @event_handler(1)
    def event_reset(self, **kwargs):
        """Handle reset control event."""
        del kwargs
        self.reset()

    def reset(self):
        """Reset the shot profile for the passed mode back to the first state (State 0) and reset all sequences."""
        self.debug_log("Resetting.")

        self.jump(state=0)

    @event_handler(2)
    def event_restart(self, **kwargs):
        """Handle restart control event."""
        del kwargs
        self.restart()

    def restart(self):
        """Restart the shot profile by calling reset() and enable().

        Automatically called when one fo the restart_events is called.
        """
        self.reset()
        self.enable()

    def _enable(self):
        super()._enable()
        self._register_switch_handlers()
        self._update_show()

    def _disable(self):
        super()._disable()
        self._remove_switch_handlers()
        self._update_show()



from mpf.tests.MpfTestCase import MpfTestCase


class TestDeviceGI(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/device/'

    def get_platform(self):
        return 'smart_virtual'

    def testBasicOnAndOff(self):
        """Tests setting some GI brightness levels (using default arguments)"""

        # Make sure hardware devices have been configured for tests
        self.assertIn('gi_01', self.machine.lights)
        self.assertIn('gi_02', self.machine.lights)

        self.advance_time_and_run(10)

        # GI should start out disabled like any light
        self.assertLightChannel("gi_01", 0)
        self.assertLightChannel("gi_02", 0)

        # Turn on GI (different brightness levels)
        self.machine.lights["gi_01"].color([128, 128, 128])
        self.advance_time_and_run(.1)
        self.assertLightChannel("gi_01", 128)
        self.machine.lights["gi_02"].color([77, 77, 77])
        self.advance_time_and_run(.1)
        self.assertLightChannel("gi_02", 77)

        # Turn off GI
        self.machine.lights["gi_01"].off()
        self.advance_time_and_run(.1)
        self.assertLightChannel("gi_01", 0)
        self.machine.lights["gi_02"].off()
        self.advance_time_and_run(.1)
        self.assertLightChannel("gi_02", 0)



"""Bcp server for clients which connect and disconnect randomly."""
from mpf.exceptions.runtime_error import MpfRuntimeError

from mpf.core.utility_functions import Util
from mpf.core.mpf_controller import MpfController


class BcpServer(MpfController):

    """Server socket which listens for incoming BCP clients."""

    config_name = "bcp_server"

    def __init__(self, machine, ip, port, server_type):
        """Initialise BCP server."""
        super().__init__(machine)
        self._server = None
        self._ip = ip
        self._port = port
        self._type = server_type

    async def start(self):
        """Start the server."""
        try:
            self._server = await self.machine.clock.start_server(
                self._accept_client, self._ip, self._port)
        except IOError as e:
            raise MpfRuntimeError("Failed to bind BCP Socket to {} on port {}. "
                                  "Is there another application running on that port?".format(self._ip, self._port), 1,
                                  "MPF BCP Server") from e

    def stop(self):
        """Stop the BCP server, i.e. closes the listening socket(s)."""
        if self._server:
            self._server.close()

        self._server = None

    async def _accept_client(self, client_reader, client_writer):
        """Accept an connection and create client."""
        self.info_log("New client connected.")
        client = Util.string_to_class(self._type)(self.machine, None, self.machine.bcp)
        client.accept_connection(client_reader, client_writer)
        client.exit_on_close = False
        self.machine.bcp.transport.register_transport(client)



"""Error in MPF or MPF-MC."""
from mpf._version import log_url


class BaseError(AssertionError):

    """Error in a config file found."""

    # pylint: disable-msg=too-many-arguments
    def __init__(self, message, error_no, logger_name, context=None, url_name=None):
        """Initialise exception."""
        self._logger_name = logger_name
        self._error_no = error_no
        self._context = context
        self._message = message
        if url_name:
            self._url_name = url_name
        else:
            self._url_name = logger_name
        super().__init__(message)

    def get_short_name(self):
        """Return short name."""
        raise NotImplementedError

    def get_long_name(self):
        """Return long name."""
        raise NotImplementedError

    def extend(self, message):
        """Chain a new message onto an existing error, keeping the original error's logger, context, and error_no."""
        self._message = "{} >> {}".format(message, self._message)
        super().__init__(self._message)

    def __str__(self):
        """Return nice string."""
        error_slug = "{}-{}-{}".format(self.get_short_name(), self._url_name.replace(" ", "_"), self._error_no)
        error_url = log_url.format(error_slug)
        if self._context:
            return "{} in {}: {} Context: {} Error Code: {} ({})".format(
                self.get_long_name(), self._logger_name, super().__str__(), self._context, error_slug, error_url)

        return "{} in {}: {} Error Code: {} ({})".format(
            self.get_long_name(), self._logger_name, super().__str__(), error_slug, error_url)



"""Fast Io board."""


class FastIoBoard:

    """A fast IO board on the NET processor."""

    # pylint: disable-msg=too-many-arguments
    def __init__(self, node_id, model_string, firmware_version, switch_count, driver_count):
        """Initialise FastIoBoard."""
        self.node_id = node_id
        self.model_string = model_string
        self.firmware_version = firmware_version
        self.switch_count = switch_count
        self.driver_count = driver_count

    def get_description_string(self) -> str:
        """Return description string."""
        return "Board {} - Model: {} Firmware: {} Switches: {} Drivers: {}".format(
            self.node_id,
            self.model_string,
            self.firmware_version,
            self.switch_count,
            self.driver_count
        )



import selectors
import socket
from asyncio import base_events, coroutine, events      # type: ignore
import collections
import heapq

# A class to manage set of next events:
from asyncio.selector_events import _SelectorSocketTransport    # type: ignore # noqa

import asyncio

import time

from mpf.core.clock import ClockBase
from serial_asyncio import SerialTransport


class NextTimers:

    """Next timers."""

    __slots__ = ["_timers_set", "_timers_heap"]

    def __init__(self):
        # Timers set. Used to check uniqueness:
        self._timers_set = set()
        # Timers heap. Used to get the closest timer event:
        self._timers_heap = []

    def add(self, when):
        """
        Add a timer (Future event).
        """
        # We don't add a time twice:
        if when in self._timers_set:
            return

        # Add to set:
        self._timers_set.add(when)
        # Add to heap:
        heapq.heappush(self._timers_heap, when)

    def is_empty(self):
        return (len(self._timers_set) == 0)

    def pop_closest(self):
        """
        Get closest event timer. (The one that will happen the soonest).
        """
        if self.is_empty():
            raise IndexError('NextTimers is empty')

        when = heapq.heappop(self._timers_heap)
        self._timers_set.remove(when)

        return when

    def __repr__(self):
        return str(self._timers_set)

class _TestTransport:

    __slots__ = ["_loop", "_sock"]

    def __init__(self, loop, sock):
        self._loop = loop
        self._sock = sock
        self._loop.add_reader(self._sock, self._read_ready)

    def _read_ready(self):
        pass

    def write(self, msg):
        pass

    def disconnect(self):
        pass


class MockFd:

    __slots__ = ["is_open"]

    def __init__(self):
        self.is_open = False

    def open(self):
        if self.is_open:
            raise AssertionError("Serial already open")
        self.is_open = True

    def read_ready(self):
        if not self.is_open:
            raise AssertionError("Serial not open")
        return False

    def send(self, data):
        if not self.is_open:
            raise AssertionError("Serial not open")
        return len(data)

    def fileno(self):
        return self

    def write_ready(self):
        if not self.is_open:
            raise AssertionError("Serial not open")
        return False

    def close(self):
        self.is_open = False
        return


class MockSocket(MockFd):

    __slots__ = ["family", "type", "proto", "__dict__"]

    def __init__(self):
        super().__init__()
        self.family = socket.AF_INET
        self.type = socket.SOCK_STREAM
        self.proto = socket.IPPROTO_TCP

    def setsockopt(self, *args, **kwargs):
        pass

    def getsockname(self):
        return ""

    def getpeername(self):
        return ""

    def recv(self, size):
        raise AssertionError("Not implemented")


class MockQueueSocket(MockSocket):

    __slots__ = ["send_queue", "recv_queue"]

    def __init__(self, loop):
        super().__init__()
        self.send_queue = asyncio.Queue()
        self.recv_queue = []

    def write_ready(self):
        return True

    def read_ready(self):
        return bool(len(self.recv_queue))

    def recv(self, size):
        return self.recv_queue.pop(0)

    def send(self, data):
        self.send_queue.put_nowait(data)
        return len(data)


class MockServer:

    __slots__ = ["loop", "is_bound", "client_connected_cb"]

    def __init__(self, loop):
        self.loop = loop
        self.is_bound = asyncio.Future()
        self.client_connected_cb = None

    async def bind(self, client_connected_cb):
        self.client_connected_cb = client_connected_cb
        self.is_bound.set_result(True)

    async def add_client(self, socket):
        if not self.is_bound.done():
            raise AssertionError("Server not running")

        limit = asyncio.streams._DEFAULT_LIMIT
        reader = asyncio.streams.StreamReader(limit=limit, loop=self.loop)
        protocol = asyncio.streams.StreamReaderProtocol(reader, loop=self.loop)
        transport = _SelectorSocketTransport(self.loop, socket, protocol)
        writer = asyncio.streams.StreamWriter(transport, protocol, reader, self.loop)
        await self.client_connected_cb(reader, writer)

    def close(self):
        pass

    async def wait_closed(self):
        return True


class MockSerial(MockFd):


    def __init__(self):
        super().__init__()
        self.fd = self
        self.timeout = None

    def reset_input_buffer(self):
        pass

    def nonblocking(self):
        pass

    def flush(self):
        pass

    @property
    def out_waiting(self):
        return 0 if self.write_ready() else 1024

    def in_waiting(self):
        return 1024 if self.read_ready() else 0

    def read(self, length):
        raise AssertionError("Not implemented")


class TestSelector(selectors.BaseSelector):

    __slots__ = ["keys"]

    def __init__(self):
        self.keys = {}

    def register(self, fileobj, events, data=None):
        key = selectors.SelectorKey(fileobj, 0, events, data)
        self.keys[fileobj] = key
        return key

    def unregister(self, fileobj):
        return self.keys.pop(fileobj)

    def select(self, timeout=None):
        del timeout
        ready = []
        for sock, key in self.keys.items():
            if sock.read_ready():
                ready.append((key, selectors.EVENT_READ))
            if sock.write_ready():
                ready.append((key, selectors.EVENT_WRITE))
        return ready

    def get_map(self):
        return self.keys


# Based on TestLoop from asyncio.test_utils:
class TimeTravelLoop(base_events.BaseEventLoop):

    """
    Loop for unittests. Passes time without waiting, but makes sure events
    happen in the correct order.
    """

    __slots__ = ["readers", "writers", "_time", "_clock_resolution", "_timers", "_selector", "_transports",
                 "_wait_for_external_executor", "_stopped"]

    def __init__(self):
        self.readers = {}
        self.writers = {}

        super().__init__()

        self._time = 0
        self._stopped = False
        self._clock_resolution = 1e-9
        self._timers = NextTimers()
        self._selector = TestSelector()
        self._transports = {}   # needed for newer asyncio on windows
        self.reset_counters()
        self._wait_for_external_executor = False

    def time(self):
        return self._time

    def set_time(self, time):
        """Set time in loop."""
        self._time = time

    def advance_time(self, advance):
        """Move test time forward."""
        if advance:
            self._time += advance

    def _add_reader(self, *args, **kwargs):
        return self.add_reader(*args, **kwargs)

    def add_reader(self, fd, callback, *args):
        """Add a reader callback."""
        self._check_closed()
        handle = events.Handle(callback, args, self)
        try:
            key = self._selector.get_key(fd)
        except KeyError:
            self._selector.register(fd, selectors.EVENT_READ,
                                    (handle, None))
        else:
            mask, (reader, writer) = key.events, key.data
            self._selector.modify(fd, mask | selectors.EVENT_READ,
                                  (handle, writer))
            if reader is not None:
                reader.cancel()

    def stop(self):
        """Stop loop."""
        self._stopped = True
        super().stop()

    def _remove_reader(self, fd):
        return self.remove_reader(fd)

    def remove_reader(self, fd):
        """Remove a reader callback."""
        if self.is_closed():
            return False
        try:
            key = self._selector.get_key(fd)
        except KeyError:
            return False
        else:
            mask, (reader, writer) = key.events, key.data
            mask &= ~selectors.EVENT_READ
            if not mask:
                self._selector.unregister(fd)
            else:
                self._selector.modify(fd, mask, (None, writer))

            if reader is not None:
                reader.cancel()
                return True
            else:
                return False

    def _add_writer(self, *args, **kwargs):
        return self.add_writer(*args, **kwargs)

    def add_writer(self, fd, callback, *args):
        """Add a writer callback.."""
        self._check_closed()
        handle = events.Handle(callback, args, self)
        try:
            key = self._selector.get_key(fd)
        except KeyError:
            self._selector.register(fd, selectors.EVENT_WRITE,
                                    (None, handle))
        else:
            mask, (reader, writer) = key.events, key.data
            self._selector.modify(fd, mask | selectors.EVENT_WRITE,
                                  (reader, handle))
            if writer is not None:
                writer.cancel()

    def _remove_writer(self, fd):
        return self.remove_writer(fd)

    def remove_writer(self, fd):
        """Remove a writer callback."""
        if self.is_closed():
            return False
        try:
            key = self._selector.get_key(fd)
        except KeyError:
            return False
        else:
            mask, (reader, writer) = key.events, key.data
            # Remove both writer and connector.
            mask &= ~selectors.EVENT_WRITE
            if not mask:
                self._selector.unregister(fd)
            else:
                self._selector.modify(fd, mask, (reader, None))

            if writer is not None:
                writer.cancel()
                return True
            else:
                return False

    def assert_writer(self, fd, callback, *args):
        assert fd in self.writers, 'fd {} is not registered'.format(fd)
        handle = self.writers[fd]
        assert handle[0] == callback, '{!r} != {!r}'.format(
            handle[0], callback)
        assert handle[1] == args, '{!r} != {!r}'.format(
            handle[1], args)

    def reset_counters(self):
        self.remove_reader_count = collections.defaultdict(int)
        self.remove_writer_count = collections.defaultdict(int)

    def run_once(self):
        if hasattr(events, "_set_running_loop"):
            events._set_running_loop(self)

        self._run_once()

        if hasattr(events, "_set_running_loop"):
            events._set_running_loop(None)

    def _run_once(self):
        # Advance time only when we finished everything at the present:
        if len(self._ready) == 0:
            if not self._timers.is_empty():
                self._time = self._timers.pop_closest()
            elif not self._closed and not self._stopped and not self._selector.select(0) and \
                    not self._wait_for_external_executor:
                raise AssertionError("Ran into an infinite loop. No socket ready and nothing scheduled.")
            if self._wait_for_external_executor:
                time.sleep(.0001)

        super()._run_once()
        if self._wait_for_external_executor:
            self._waiting_since = None

    def call_at(self, when, callback, *args, **kwargs):
        self._timers.add(when)
        return super().call_at(when, callback, *args, **kwargs)

    def _process_events(self, event_list):
        for key, mask in event_list:
            fileobj, (reader, writer) = key.fileobj, key.data
            if mask & selectors.EVENT_READ and reader is not None:
                if reader._cancelled:
                    self.remove_reader(fileobj)
                else:
                    self._add_callback(reader)
            if mask & selectors.EVENT_WRITE and writer is not None:
                if writer._cancelled:
                    self.remove_writer(fileobj)
                else:
                    self._add_callback(writer)

    def _write_to_self(self):
        pass


class TestClock(ClockBase):

    __slots__ = ["_test_loop", "_mock_sockets", "_mock_servers", "_mock_serials"]

    def __init__(self, loop):
        self._test_loop = loop
        super().__init__()
        self._mock_sockets = {}
        self._mock_servers = {}
        self._mock_serials = {}

    def _create_event_loop(self):
        return self._test_loop

    def mock_socket(self, host, port, socket):
        """Mock a socket and use it for connections."""
        self._mock_sockets[host + ":" + str(port)] = socket

    def mock_server(self, host, port, server):
        """Mock a server and use it for connections."""
        self._mock_servers[host + ":" + str(port)] = server

    def _open_mock_socket(self, host, port):
        key = host + ":" + str(port)
        if key not in self._mock_sockets:
            raise AssertionError("socket not mocked for key {}".format(key))
        socket = self._mock_sockets[key]
        if socket.is_open:
            raise AssertionError("socket already open for key {}".format(key))

        socket.is_open = True
        return socket

    async def start_server(self, client_connected_cb, host=None, port=None, **kwd):
        """Mock listening server."""
        key = host + ":" + str(port)
        if key not in self._mock_servers:
            raise AssertionError("server not mocked for key {}".format(key))
        server = self._mock_servers[key]
        if server.is_bound.done():
            raise AssertionError("server already bound for key {}".format(key))

        await server.bind(client_connected_cb)
        return server

    @coroutine
    def open_connection(self, host=None, port=None, *,
                        limit=None, **kwds):
        """A wrapper for create_connection() returning a (reader, writer) pair.

        The reader returned is a StreamReader instance; the writer is a
        StreamWriter instance.

        The arguments are all the usual arguments to create_connection()
        except protocol_factory; most common are positional host and port,
        with various optional keyword arguments following.

        Additional optional keyword arguments are loop (to set the event loop
        instance to use) and limit (to set the buffer limit passed to the
        StreamReader).

        (If you want to customize the StreamReader and/or
        StreamReaderProtocol classes, just copy the code -- there's
        really nothing special here except some convenience.)
        """
        if not limit:
            limit = asyncio.streams._DEFAULT_LIMIT  # type: ignore
        reader = asyncio.streams.StreamReader(limit=limit, loop=self.loop)
        protocol = asyncio.streams.StreamReaderProtocol(reader, loop=self.loop)
        sock = self._open_mock_socket(host, port)
        transport = _SelectorSocketTransport(self.loop, sock, protocol)
        writer = asyncio.streams.StreamWriter(transport, protocol, reader, self.loop)
        return reader, writer   # type: ignore

    def mock_serial(self, url, serial):
        """Mock a socket and use it for connections."""
        self._mock_serials[url] = serial

    def _open_mock_serial(self, url, do_not_open):
        key = url
        if key not in self._mock_serials:
            raise AssertionError("serial not mocked for key {}".format(key))
        serial = self._mock_serials[key]
        if not do_not_open:
            if serial.is_open:
                raise AssertionError("serial already open for key {}".format(key))

            serial.is_open = True
        return serial

    @coroutine
    def open_serial_connection(self, limit=None, **kwargs):     # type: ignore
        """A wrapper for create_serial_connection() returning a (reader,
        writer) pair.

        The reader returned is a StreamReader instance; the writer is a
        StreamWriter instance.

        The arguments are all the usual arguments to Serial(). Additional
        optional keyword arguments are loop (to set the event loop instance
        to use) and limit (to set the buffer limit passed to the
        StreamReader.

        This function is a coroutine.
        """
        if not limit:
            limit = asyncio.streams._DEFAULT_LIMIT  # type: ignore

        reader = asyncio.StreamReader(limit=limit, loop=self.loop)
        protocol = asyncio.StreamReaderProtocol(reader, loop=self.loop)
        transport = SerialTransport(self.loop, protocol, self._open_mock_serial(kwargs['url'],
                                                                                kwargs.get("do_not_open", False)))
        writer = asyncio.StreamWriter(transport, protocol, reader, self.loop)
        return reader, writer



"""Contains the Mode base class."""
from typing import Any, Optional
from typing import Callable
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

from mpf.core.delays import DelayManager
from mpf.core.logging import LogMixin
from mpf.core.switch_controller import SwitchHandler
from mpf.core.events import EventHandlerKey
from mpf.core.events import QueuedEvent  # pylint: disable-msg=cyclic-import,unused-import

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.mode_device import ModeDevice     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.player import Player  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


# pylint: disable-msg=too-many-instance-attributes
class Mode(LogMixin):

    """Base class for a mode."""

    __slots__ = ["machine", "config", "name", "path", "priority", "_active", "_starting", "_mode_start_wait_queue",
                 "stop_methods", "start_callback", "stop_callbacks", "event_handlers", "switch_handlers",
                 "mode_stop_kwargs", "mode_devices", "start_event_kwargs", "stopping", "delay", "player",
                 "auto_stop_on_ball_end", "restart_on_next_ball", "asset_paths"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, machine: "MachineController", config, name: str, path, asset_paths) -> None:
        """Initialise mode.

        Args:
        ----
            machine: the machine controller
            config: config dict for mode
            name: name of mode
            path: path of mode
            asset_paths: all paths to consider for assets in this mode
        """
        super().__init__()
        self.machine = machine                  # type: MachineController
        self.config = config                    # type: ignore
        self.name = name
        self.path = path
        self.asset_paths = asset_paths
        self.priority = 0
        self._active = False
        self._starting = False
        self._mode_start_wait_queue = None      # type: Optional[QueuedEvent]
        self.stop_methods = list()              # type: List[Tuple[Callable[[Any], None], Any]]
        self.start_callback = None              # type: Optional[Callable[[], None]]
        self.stop_callbacks = []                # type: List[Callable[[], None]]
        self.event_handlers = set()             # type: Set[EventHandlerKey]
        self.switch_handlers = list()           # type: List[SwitchHandler]
        self.mode_stop_kwargs = dict()          # type: Dict[str, Any]
        self.mode_devices = set()               # type: Set[ModeDevice]
        self.start_event_kwargs = {}            # type: Dict[str, Any]
        self.stopping = False

        self.delay = DelayManager(self.machine)
        '''DelayManager instance for delays in this mode. Note that all delays
        scheduled here will be automatically canceled when the mode stops.'''

        self.player = None                      # type: Optional[Player]
        '''Reference to the current player object.'''

        self.configure_logging('Mode.' + name,
                               self.config['mode']['console_log'],
                               self.config['mode']['file_log'])

        self.configure_mode_settings(config.get('mode', dict()))

        self.auto_stop_on_ball_end = self.config['mode']['stop_on_ball_end']
        '''Controls whether this mode is stopped when the ball ends,
        regardless of its stop_events settings.
        '''

        self.restart_on_next_ball = self.config['mode']['restart_on_next_ball']
        '''Controls whether this mode will restart on the next ball. This only
        works if the mode was running when the ball ended. It's tracked per-
        player in the 'restart_modes_on_next_ball' player variable.
        '''

        if self.config['mode']['game_mode'] and not self.config['mode']['stop_on_ball_end']:
            self.raise_config_error("All game modes need to stop at ball end. If you want to set stop_on_ball_end to "
                                    "False also set game_mode to False.", 1)

    @staticmethod
    def get_config_spec() -> str:
        """Return config spec for mode_settings."""
        return '''
                __valid_in__: mode
                __allow_others__:
                '''

    def __repr__(self):
        """Return string representation."""
        return '<Mode.{}>'.format(self.name)

    @property
    def active(self) -> bool:
        """Return *True* if this mode is active."""
        return self._active

    @active.setter
    def active(self, new_active: bool):
        """Setter for _active."""
        if self._active != new_active:
            self._active = new_active
            self.machine.mode_controller.set_mode_state(self, self._active)

    def configure_mode_settings(self, config: dict) -> None:
        """Process this mode's configuration settings from a config dictionary."""
        self.config['mode'] = self.machine.config_validator.validate_config(
            config_spec='mode', source=config, section_name='mode')

        for event in self.config['mode']['start_events']:
            self.machine.events.add_handler(event=event, handler=self.start,
                                            priority=self.config['mode']['priority'] +
                                            self.config['mode']['start_priority'])

    @property
    def is_game_mode(self) -> bool:
        """Return true if this is a game mode."""
        return bool(self.config['mode']['game_mode'])

    def start(self, mode_priority=None, callback=None, **kwargs) -> None:
        """Start this mode.

        Args:
        ----
            mode_priority: Integer value of what you want this mode to run at. If you
                don't specify one, it will use the "Mode: priority" setting from
                this mode's configuration file.
            callback: Callback to call when this mode has been started.
            **kwargs: Catch-all since this mode might start from events with
                who-knows-what keyword arguments.

        Warning: You can safely call this method, but do not override it in your
        mode code. If you want to write your own mode code by subclassing Mode,
        put whatever code you want to run when this mode starts in the
        mode_start method which will be called automatically.
        """
        # remove argument so we do not repost this
        kwargs.pop('_from_bcp', None)
        self.debug_log("Received request to start")

        if self.config['mode']['game_mode'] and not (self.machine.game and self.player):
            self.warning_log("Can only start mode %s during a game. Aborting start.", self.name)
            return

        if self._active:
            self.debug_log("Mode is already active. Aborting start.")
            return

        if self._starting:
            self.debug_log("Mode already starting. Aborting start.")
            return

        self._starting = True

        self.machine.events.post('mode_{}_will_start'.format(self.name), **kwargs)
        '''event: mode_(name)_will_start

        desc: Posted when a mode is about to start. The "name" part is replaced
        with the actual name of the mode, so the actual event posted is
        something like *mode_attract_will_start*, *mode_base_will_start*, etc.

        This is posted before the "mode_(name)_starting" event.
        '''

        if self.config['mode']['use_wait_queue'] and 'queue' in kwargs:

            self.debug_log("Registering a mode start wait queue")

            self._mode_start_wait_queue = kwargs['queue']
            assert isinstance(self._mode_start_wait_queue, QueuedEvent)
            self._mode_start_wait_queue.wait()

        if isinstance(mode_priority, int):
            self.priority = mode_priority
        else:
            self.priority = self.config['mode']['priority']

        self.start_event_kwargs = kwargs

        # hook for custom code. called before any mode devices are set up
        self.mode_will_start(**self.start_event_kwargs)

        self._add_mode_devices()

        self.debug_log("Registering mode_stop handlers")

        # register mode stop events
        if 'stop_events' in self.config['mode']:

            for event in self.config['mode']['stop_events']:
                # stop priority is +1 so if two modes of the same priority
                # start and stop on the same event, the one will stop before
                # the other starts
                self.add_mode_event_handler(event=event, handler=self.stop,
                                            priority=self.config['mode']['stop_priority'] + 1)

        self.start_callback = callback

        self.debug_log("Calling mode_start handlers")

        for item in self.machine.mode_controller.start_methods:
            if item.config_section in self.config or not item.config_section:
                result = item.method(config=self.config.get(item.config_section, self.config),
                                     priority=self.priority,
                                     mode=self,
                                     **item.kwargs)
                if result:
                    self.stop_methods.append(result)

        self._setup_device_control_events()

        self.machine.events.post_queue(event='mode_{}_starting'.format(self.name),
                                       callback=self._started, **kwargs)
        '''event: mode_(name)_starting

        desc: The mode called "name" is starting.

        This is a queue event. The mode will not fully start until the queue is
        cleared.
        '''

    def _started(self, **kwargs) -> None:
        """Handle result of mode_<name>_starting queue event."""
        del kwargs
        if self.machine.is_shutting_down:
            self.info_log("Will not start because machine is shutting down.")
            return

        self.info_log('Started. Priority: %s', self.priority)

        self.active = True
        self._starting = False

        for event_name in self.config['mode']['events_when_started']:
            self.machine.events.post(event_name)

        self.machine.events.post(event='mode_{}_started'.format(self.name), callback=self._mode_started_callback,
                                 **self.start_event_kwargs)
        '''event: mode_(name)_started

        desc: Posted when a mode has started. The "name" part is replaced
        with the actual name of the mode, so the actual event posted is
        something like *mode_attract_started*, *mode_base_started*, etc.

        This is posted after the "mode_(name)_starting" event.
        '''

    def _mode_started_callback(self, **kwargs) -> None:
        """Handle result of mode_<name>_started queue event."""
        del kwargs
        self.mode_start(**self.start_event_kwargs)

        self.start_event_kwargs = dict()

        if self.start_callback:
            self.start_callback()

        self.debug_log('Mode Start process complete.')

    def stop(self, callback: Any = None, **kwargs) -> bool:
        """Stop this mode.

        Args:
        ----
            callback: Method which will be called once this mode has stopped. Will only be called when the mode is
                running (includes currently stopping)
            **kwargs: Catch-all since this mode might start from events with
                who-knows-what keyword arguments.

        Warning: You can safely call this method, but do not override it in your
        mode code. If you want to write your own mode code by subclassing Mode,
        put whatever code you want to run when this mode stops in the
        mode_stop method which will be called automatically.

        Returns true if the mode is running. Otherwise false.
        """
        if not self._active:
            return False

        if callback:
            self.stop_callbacks.append(callback)

        # do not stop twice. only register callback in that case
        if self.stopping:
            # mode is still running
            return True

        self.machine.events.post('mode_' + self.name + '_will_stop')
        '''event: mode_(name)_will_stop

        desc: Posted when a mode is about to stop. The "name" part is replaced
        with the actual name of the mode, so the actual event posted is
        something like *mode_attract_will_stop*, *mode_base_will_stop*, etc.

        This is posted immediately before the "mode_(name)_stopping" event.
        '''

        self.stopping = True

        self.mode_stop_kwargs = kwargs

        self.debug_log('Mode Stopping.')

        self._remove_mode_switch_handlers()

        self.delay.clear()

        self.machine.events.post_queue(event='mode_' + self.name + '_stopping',
                                       callback=self._stopped)
        '''event: mode_(name)_stopping

        desc: The mode called "name" is stopping. This is a queue event. The
        mode won't actually stop until the queue is cleared.

        '''
        return True

    def _stopped(self) -> None:
        self.info_log('Stopped.')

        self.priority = 0
        self.active = False
        self.stopping = False

        for item in self.stop_methods:
            item[0](item[1])

        self.stop_methods = list()

        for event_name in self.config['mode']['events_when_stopped']:
            self.machine.events.post(event_name)

        self.machine.events.post('mode_' + self.name + '_stopped',
                                 callback=self._mode_stopped_callback)
        '''event: mode_(name)_stopped

        desc: Posted when a mode has stopped. The "name" part is replaced
        with the actual name of the mode, so the actual event posted is
        something like *mode_attract_stopped*, *mode_base_stopped*, etc.
        '''

        self.machine.events.post('clear', key=self.name)
        '''event: clear

        args:
            key: string name of the configs to clear

        desc: Posted to cause config players to clear whatever they're running
            based on the key passed. Typically posted when a show or mode ends.
        '''

        if self._mode_start_wait_queue:

            self.debug_log("Clearing wait queue")

            self._mode_start_wait_queue.clear()
            self._mode_start_wait_queue = None

    def _mode_stopped_callback(self, **kwargs) -> None:
        del kwargs

        # Call the mode_stop() method before removing the devices
        self.mode_stop(**self.mode_stop_kwargs)
        self.mode_stop_kwargs = dict()

        # Clean up the mode handlers and devices
        self._remove_mode_event_handlers()
        self._remove_mode_devices()

        for callback in self.stop_callbacks:
            callback()

        self.stop_callbacks = []

    def _add_mode_devices(self) -> None:
        # adds and initializes mode devices which get removed at the end of the mode

        for collection_name, device_class in (
                iter(self.machine.device_manager.device_classes.items())):

            # check if there is config for the device type
            if device_class.config_section in self.config:

                for device_name in self.config[device_class.config_section]:

                    collection = getattr(self.machine, collection_name)

                    # get device
                    device = collection[device_name]

                    # Track that this device was added via this mode so we
                    # can remove it when the mode ends.
                    self.mode_devices.add(device)
                    if not self.config['mode']['game_mode'] and not device.can_exist_outside_of_game:
                        raise AssertionError("Device {} cannot exist in non game-mode {}.".format(
                            device, self.name
                        ))

                    # This lets the device know it was added to a mode
                    device.device_loaded_in_mode(mode=self, player=self.player)

    def create_mode_devices(self) -> None:
        """Create new devices that are specified in a mode config that haven't been created in the machine-wide."""
        self.debug_log("Scanning config for mode-based devices")

        for collection_name, device_class in iter(self.machine.device_manager.device_classes.items()):

            # check if there is config for the device type
            if device_class.config_section not in self.config:
                continue

            for device, settings in iter(self.config[device_class.config_section].items()):

                collection = getattr(self.machine, collection_name)

                if device not in collection:  # no existing device, create

                    self.debug_log("Creating mode-based device: %s",
                                   device)

                    self.machine.device_manager.create_devices(
                        collection.name, {device: settings})

    async def load_mode_devices(self) -> None:
        """Load config of mode devices."""
        for collection_name, device_class in iter(self.machine.device_manager.device_classes.items()):

            # check if there is config for the device type
            if device_class.config_section not in self.config:
                continue

            for device, settings in iter(self.config[device_class.config_section].items()):

                collection = getattr(self.machine, collection_name)
                device = collection[device]
                settings = device.prepare_config(settings, True)
                settings = device.validate_and_parse_config(settings, True, "mode:" + self.name)

                if device.config:
                    self.debug_log("Overwrite mode-based device: %s", device)
                    # overload
                    device.overload_config_in_mode(self, settings)

                else:
                    self.debug_log("Initializing mode-based device: %s", device)
                    # load config
                    device.load_config(settings)

        for collection_name, device_class in iter(self.machine.device_manager.device_classes.items()):
            # check if there is config for the device type
            if device_class.config_section not in self.config:
                continue

            for device, settings in iter(self.config[device_class.config_section].items()):
                collection = getattr(self.machine, collection_name)
                device = collection[device]
                await device.device_added_to_mode(mode=self)

    def _remove_mode_devices(self) -> None:
        for device in self.mode_devices:
            device.device_removed_from_mode(self)

        self.mode_devices = set()

    def _setup_device_control_events(self) -> None:
        # registers mode handlers for control events for all devices specified
        # in this mode's config (not just newly-created devices)

        self.debug_log("Scanning mode-based config for device control_events")

        for event, method, delay, device in (
                self.machine.device_manager.get_device_control_events(
                self.config)):

            if not delay:
                self.add_mode_event_handler(
                    event=event,
                    handler=method,
                    blocking_facility=device.class_label)
            else:
                self.add_mode_event_handler(
                    event=event,
                    handler=self._control_event_handler,
                    callback=method,
                    ms_delay=delay,
                    blocking_facility=device.class_label)

        # get all devices in the mode
        device_list = set()     # type: Set[ModeDevice]
        for collection in self.machine.device_manager.collections:
            if self.machine.device_manager.collections[collection].config_section in self.config:
                for device, _ in \
                        iter(self.config[self.machine.device_manager.collections[collection].config_section].items()):
                    device_list.add(self.machine.device_manager.collections[collection][device])

        for device in device_list:
            device.add_control_events_in_mode(self)

    def _control_event_handler(self, callback: Callable[..., None], ms_delay: int = 0, **kwargs) -> None:
        del kwargs
        self.debug_log("_control_event_handler: callback: %s,", callback)

        self.delay.add(ms=ms_delay, callback=callback, mode=self)

    def add_mode_event_handler(self, event: str, handler: Callable, priority: int = 0, **kwargs) -> EventHandlerKey:
        """Register an event handler which is automatically removed when this mode stops.

        This method is similar to the Event Manager's add_handler() method,
        except this method automatically unregisters the handlers when the mode
        ends.

        Args:
        ----
            event: String name of the event you're adding a handler for. Since
                events are text strings, they don't have to be pre-defined.
            handler: The method that will be called when the event is fired.
            priority: An arbitrary integer value that defines what order the
                handlers will be called in. The default is 1, so if you have a
                handler that you want to be called first, add it here with a
                priority of 2. (Or 3 or 10 or 100000.) The numbers don't matter.
                They're called from highest to lowest. (i.e. priority 100 is
                called before priority 1.)
            **kwargs: Any any additional keyword/argument pairs entered here
                will be attached to the handler and called whenever that handler
                is called. Note these are in addition to kwargs that could be
                passed as part of the event post. If there's a conflict, the
                event-level ones will win.

        Returns a EventHandlerKey to the handler which you can use to later remove
        the handler via ``remove_handler_by_key``. Though you don't need to
        remove the handler since the whole point of this method is they're
        automatically removed when the mode stops.

        Note that if you do add a handler via this method and then remove it
        manually, that's ok too.
        """
        key = self.machine.events.add_handler(event, handler, self.priority + priority, mode=self, **kwargs)

        self.event_handlers.add(key)

        return key

    def _remove_mode_event_handlers(self) -> None:
        for key in self.event_handlers:
            self.machine.events.remove_handler_by_key(key)
        self.event_handlers = set()

    def _remove_mode_switch_handlers(self) -> None:
        for handler in self.switch_handlers:
            self.machine.switch_controller.remove_switch_handler_by_key(handler)
        self.switch_handlers = list()

    def initialise_mode(self) -> None:
        """Initialise this mode."""
        self.mode_init()

    def mode_init(self) -> None:
        """User-overrideable method which will be called when this mode initializes as part of the MPF boot process."""

    def mode_will_start(self, **kwargs) -> None:
        """User-overrideable method which will be called whenever this mode starts (i.e. before it becomes active)."""

    def mode_start(self, **kwargs) -> None:
        """User-overrideable method which will be called whenever this mode starts (i.e. whenever it becomes active)."""

    def mode_stop(self, **kwargs) -> None:
        """User-overrideable method which will be called whenever this mode stops."""
